from textwrap import dedent

import click

from glotaran.plugin_system.data_io_registration import known_data_formats
from glotaran.plugin_system.megacomplex_registration import known_megacomplex_names
from glotaran.plugin_system.project_io_registration import known_project_formats


def plugin_list_cmd():
    """Prints a list of installed plugins."""

    output = dedent(
        """
        Installed Glotaran Plugins:

        Megacomplex Models:
        """
    )
    output += "\n"

    for name in known_megacomplex_names():
        output += f"    * {name}\n"

    output += "\nData file Formats\n\n"

    for reader_fmt in known_data_formats():
        output += f"    * .{reader_fmt}\n"

    output += "\nProject file Formats\n\n"

    for reader_fmt in known_project_formats():
        output += f"    * .{reader_fmt}\n"

    click.echo(output)



"""Module for residual calculation with the non-negative least-squares method."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.optimize import nnls

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


def residual_nnls(matrix: ArrayLike, data: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
    """Calculate the conditionally linear parameters and residual with the NNLS method.

    NNLS stands for 'non-negative least-squares'.

    Parameters
    ----------
    matrix : ArrayLike
        The model matrix.
    data : ArrayLike
        The data to analyze.

    Returns
    -------
    tuple[ArrayLike, ArrayLike]
        The clps and the residual.
    """
    clp, _ = nnls(matrix, data)
    residual = data - np.dot(matrix, clp)
    return clp, residual



"""Contains helper methods for dataclasses."""
from __future__ import annotations

from collections.abc import Mapping
from collections.abc import Sequence
from dataclasses import MISSING
from dataclasses import field
from dataclasses import fields
from dataclasses import is_dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from glotaran.utils.io import relative_posix_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any
    from typing import TypeVar

    from _typeshed import DataclassInstance

    from glotaran.typing.protocols import FileLoadable

    DefaultType = TypeVar("DefaultType")
    DataclassInstanceType = TypeVar("DataclassInstanceType", bound=DataclassInstance)


def exclude_from_dict_field(
    default: DefaultType = MISSING,  # type:ignore[assignment]
) -> DefaultType:
    """Create a dataclass field with which will be excluded from ``asdict``.

    Parameters
    ----------
    default : DefaultType
        The default value of the field.

    Returns
    -------
    DefaultType
        The created field.
    """
    return field(default=default, metadata={"exclude_from_dict": True})


def file_loader_factory(
    targetClass: type[FileLoadable], *, is_wrapper_class: bool = False
) -> Callable[[FileLoadable | str | Path], FileLoadable]:
    """Create ``file_loader`` functions to load ``targetClass`` from file.

    Parameters
    ----------
    targetClass: type[FileLoadable]
        Class the loader function should return an instance of.
    is_wrapper_class: bool
        Whether or not ``targetClass`` is a wrapper class, so the isinstance check will be ignored
        and instead the responsibility for supported types lies at the implementation of
        the loader.

    Returns
    -------
    file_loader: Callable[[FileLoadable | str | Path], FileLoadable]
        Function to load ``FileLoadable`` from source file or return instance if already loaded.
    """

    def file_loader(
        source_path: FileLoadable | str | Path, folder: str | Path | None = None
    ) -> FileLoadable:
        """Functions to load ``targetClass`` from file.

        Parameters
        ----------
        source_path : FileLoadable | str | Path
            Instance to ``targetClass`` or a file path to load it from.
        folder : str | Path | None
            Path to the base folder ``source_path`` is a relative path to., by default None

        Returns
        -------
        FileLoadable
            Instance of ``targetClass``.

        Raises
        ------
        ValueError
            If not an instance of ``targetClass`` or a source path to load from.
        """
        if isinstance(source_path, targetClass):
            return source_path
        if isinstance(source_path, (str, Path)):
            if folder is not None:
                target_obj = targetClass.loader(Path(folder) / source_path)
            else:
                target_obj = targetClass.loader(source_path)
            target_obj.source_path = str(source_path)
            return target_obj  # type:ignore[return-value]
        if is_wrapper_class is True:
            if isinstance(source_path, Sequence) and folder is not None:
                source_path = [Path(folder) / val for val in source_path]
            if isinstance(source_path, Mapping) and folder is not None:
                source_path = {key: Path(folder) / val for key, val in source_path.items()}
            return targetClass.loader(source_path)  # type:ignore[return-value, arg-type]
        raise ValueError(
            f"The value of 'source_path' needs to be of class {targetClass.__name__} "
            "or a file path. If the class is a wrapper class, you can use the argument:\n"
            "'is_wrapper_class=True'"
        )

    return file_loader


def file_loadable_field(
    targetClass: type[FileLoadable], *, is_wrapper_class=False
) -> FileLoadable:
    """Create a dataclass field which can be and object of type ``targetClass`` or file path.

    Parameters
    ----------
    targetClass : type[FileLoadable]
        Class the resulting value should be an instance of.
    is_wrapper_class: bool
        Whether or not ``targetClass`` is a wrapper class, so the isinstance check will be ignored
        and instead the responsibility for supported types lies at the implementation of
        the loader.

    Notes
    -----
    This also requires to add ``init_file_loadable_fields`` in the ``__post_init__`` method.

    Returns
    -------
    FileLoadable
        Instance of ``targetClass``.

    See Also
    --------
    init_file_loadable_fields
    """
    return field(
        metadata={
            "file_loader": file_loader_factory(targetClass, is_wrapper_class=is_wrapper_class)
        }
    )


def init_file_loadable_fields(dataclass_instance: DataclassInstance):
    """Load objects into class when dataclass is initialized with paths.

    If the class has file_loadable fields, this needs be called in the
    ``__post_init__`` method of that class.

    Parameters
    ----------
    dataclass_instance : DataclassInstance
        Instance of the dataclass being initialized.
        When used inside of ``__post_init__`` for the class itself use ``self``.

    See Also
    --------
    file_loadable_field
    """
    for field_item in fields(dataclass_instance):
        if "file_loader" in field_item.metadata:
            file_loader = field_item.metadata["file_loader"]
            value = getattr(dataclass_instance, field_item.name)
            setattr(dataclass_instance, field_item.name, file_loader(value))


def asdict(dataclass: DataclassInstance, folder: Path | None = None) -> dict[str, Any]:
    """Create a dictionary containing all fields of the dataclass.

    Parameters
    ----------
    dataclass : DataclassInstance
        A dataclass instance.
    folder: Path | None
        Parent folder of :class:`FileLoadable` fields. by default None

    Returns
    -------
    dict[str, Any] :
        The dataclass represented as a dictionary.
    """
    dataclass_dict = {}
    for field_item in fields(dataclass):
        if "exclude_from_dict" not in field_item.metadata:
            value = getattr(dataclass, field_item.name)
            dataclass_dict[field_item.name] = asdict(value) if is_dataclass(value) else value
        if "file_loader" in field_item.metadata:
            value = getattr(dataclass, field_item.name)
            if value.source_path is not None:
                if isinstance(value.source_path, (str, Path)):
                    dataclass_dict[field_item.name] = relative_posix_path(
                        value.source_path, folder
                    )
                elif isinstance(value.source_path, Sequence):
                    dataclass_dict[field_item.name] = [
                        relative_posix_path(val, folder) for val in value.source_path
                    ]
                elif isinstance(value.source_path, Mapping):
                    dataclass_dict[field_item.name] = {
                        key: relative_posix_path(val, folder)
                        for key, val in value.source_path.items()
                    }

    return dataclass_dict


def fromdict(
    dataclass_type: type[DataclassInstanceType],
    dataclass_dict: dict[str, Any],
    folder: Path | None = None,
) -> DataclassInstanceType:
    """Create a dataclass instance from a dict and loads all file represented fields.

    Parameters
    ----------
    dataclass_type : type[DataclassInstanceType]
        A dataclass type.
    dataclass_dict : dict[str, Any]
        A dict for instancing the the dataclass.
    folder : Path
        The root folder for file paths. If ``None`` file paths are consider absolute.

    Returns
    -------
    DataclassInstanceType
        Created instance of dataclass_type.
    """
    for field_item in fields(dataclass_type):
        if "file_loader" in field_item.metadata:
            file_path = dataclass_dict.get(field_item.name)
            dataclass_dict[field_item.name] = field_item.metadata["file_loader"](file_path, folder)
        elif is_dataclass(field_item.default) and field_item.name in dataclass_dict:
            dataclass_dict[field_item.name] = type(field_item.default)(  # type:ignore[misc]
                **dataclass_dict[field_item.name]
            )

    return dataclass_type(**dataclass_dict)



"""Utility module for ``glotaran.builtin.io.yml.yml``."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

if TYPE_CHECKING:
    from collections.abc import Mapping
    from collections.abc import Sequence
    from typing import Any

    from ruamel.yaml.nodes import ScalarNode
    from ruamel.yaml.representer import BaseRepresenter


def write_dict(
    data: Mapping[str, Any] | Sequence[Any], file_name: str | Path | None = None, offset: int = 0
) -> str | None:
    """Write a mapping (e.g. ``dict``) or sequence (e.g. ``list``) as ``yaml`` to file or str.

    Parameters
    ----------
    data: Mapping[str, Any] | Sequence[Any]
        Data that should be converted to ``yaml``.
    file_name: str | Path | None
        Path of the file to write the ``yaml`` code to.
        Defaults to None which makes this function return a string.
    offset: int
        Block indentation level. Defaults to 0
        See https://yaml.readthedocs.io/en/latest/detail.html#indentation-of-block-sequences

    Returns
    -------
    str | None
        String if ``file_name`` is ``None`` or ``None`` if ``file_name`` is a valid path.
    """
    yaml = YAML()
    yaml.representer.add_representer(type(None), _yaml_none_representer)
    yaml.indent(mapping=2, sequence=2, offset=offset)

    if file_name is not None:
        with open(file_name, "w", encoding="utf8") as f:
            yaml.dump(data, f)
    else:
        stream = StringIO()
        yaml.dump(data, stream)
        return stream.getvalue()
    return None


def load_dict(source: str | Path, is_file: bool) -> dict[str, Any]:
    """Load ``yaml`` code from a file or string and returns the dict interpretation.

    Parameters
    ----------
    source: str | Path
        Path to a file or string containing the ``yaml`` code.
    is_file: bool
        Whether or not ``source`` is a file.

    Returns
    -------
    dict[str, Any]
    """
    yaml = YAML()
    yaml.representer.add_representer(type(None), _yaml_none_representer)
    if is_file:
        with open(source) as f:
            spec = yaml.load(f)
    else:
        spec = yaml.load(source)
    return spec


def _yaml_none_representer(representer: BaseRepresenter, data: Mapping[str, Any]) -> ScalarNode:
    """Yaml repr for ``None`` python values.

    Parameters
    ----------
    representer : BaseRepresenter
        Representer of the :class:`YAML` instance.
    data : Mapping[str, Any]
        Data to write to yaml.

    Returns
    -------
    ScalarNode
        Node representing the value.

    References
    ----------
    https://stackoverflow.com/a/44314840
    """
    return representer.represent_scalar("tag:yaml.org,2002:null", "null")



from __future__ import annotations

from typing import TYPE_CHECKING

from glotaran.io import load_model
from glotaran.io import save_model
from glotaran.testing.simulated_data.sequential_spectral_decay import MODEL

if TYPE_CHECKING:
    from pathlib import Path


want = """\
clp_penalties: []
clp_constraints: []
clp_relations: []
dataset_groups:
  default:
    label: default
    residual_function: variable_projection
    link_clp: null
weights: []
megacomplex:
  megacomplex_sequential_decay:
    label: megacomplex_sequential_decay
    dimension: time
    compartments:
    - species_1
    - species_2
    - species_3
    rates:
    - rates.species_1
    - rates.species_2
    - rates.species_3
    type: decay-sequential
irf:
  gaussian_irf:
    label: gaussian_irf
    scale: null
    shift: null
    normalize: true
    backsweep: false
    backsweep_period: null
    type: gaussian
    center: irf.center
    width: irf.width
dataset:
  dataset_1:
    label: dataset_1
    group: default
    force_index_dependent: false
    megacomplex:
    - megacomplex_sequential_decay
    megacomplex_scale: null
    global_megacomplex: null
    global_megacomplex_scale: null
    scale: null
    irf: gaussian_irf
"""


def test_save_model(
    tmp_path: Path,
):
    """Check all files exist."""

    model_path = tmp_path / "testmodel.yml"
    save_model(file_name=model_path, format_name="yml", model=MODEL)

    assert model_path.is_file()
    assert model_path.read_text() == want
    assert load_model(model_path).valid()



from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.builtin.megacomplexes.decay import DecayParallelMegacomplex
from glotaran.builtin.megacomplexes.decay import DecaySequentialMegacomplex
from glotaran.model import Model
from glotaran.optimization.optimize import optimize
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation import simulate


def create_gaussian_clp(labels, amplitudes, centers, widths, axis):
    return xr.DataArray(
        [
            amplitudes[i] * np.exp(-np.log(2) * np.square(2 * (axis - centers[i]) / widths[i]))
            for i, _ in enumerate(labels)
        ],
        coords=[("clp_label", labels), ("pixel", axis.data)],
    ).T


DecaySimpleModel = Model.create_class_from_megacomplexes(
    [DecayParallelMegacomplex, DecaySequentialMegacomplex]
)
DecayModel = Model.create_class_from_megacomplexes([DecayMegacomplex])


class OneComponentOneChannel:
    model = DecayModel(
        **{
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["2"]},
            },
            "megacomplex": {
                "mc1": {"type": "decay", "k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = Parameters.from_list(
        [101e-4, [1, {"vary": False, "non-negative": False}]]
    )
    wanted_parameters = Parameters.from_list([101e-3, [1, {"vary": False, "non-negative": False}]])

    time = np.arange(0, 50, 1.5)
    pixel = np.asarray([0])
    axis = {"time": time, "pixel": pixel}

    clp = xr.DataArray([[1]], coords=[("pixel", pixel.data), ("clp_label", ["s1"])])


class OneComponentOneChannelGaussianIrf:
    model = DecayModel(
        **{
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["5"]},
            },
            "megacomplex": {
                "mc1": {"type": "decay", "k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "irf": {
                "irf1": {"type": "gaussian", "center": "2", "width": "3", "shift": ["4"]},
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = Parameters.from_list(
        [
            101e-4,
            0.1,
            1,
            [0.1, {"vary": False}],
            [1, {"vary": False, "non-negative": False}],
        ]
    )
    print(initial_parameters)
    wanted_parameters = Parameters.from_list(
        [
            [101e-3, {"non-negative": True}],
            [0.2, {"non-negative": True}],
            [2, {"non-negative": True}],
            [0.1, {"vary": False}],
            [1, {"vary": False, "non-negative": False}],
        ]
    )

    time = np.arange(0, 50, 1.5)
    pixel = np.asarray([0])
    axis = {"time": time, "pixel": pixel}

    clp = xr.DataArray([[1]], coords=[("pixel", pixel.data), ("clp_label", ["s1"])])


class ThreeComponentParallel:
    model = DecaySimpleModel(
        **{
            "megacomplex": {
                "mc1": {
                    "type": "decay-parallel",
                    "compartments": ["s1", "s2", "s3"],
                    "rates": [
                        "kinetic.1",
                        "kinetic.2",
                        "kinetic.3",
                    ],
                },
            },
            "irf": {
                "irf1": {
                    "type": "multi-gaussian",
                    "center": ["irf.center"],
                    "width": ["irf.width"],
                },
            },
            "dataset": {
                "dataset1": {
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = Parameters.from_dict(
        {
            "kinetic": [
                ["1", 501e-3],
                ["2", 202e-4],
                ["3", 105e-5],
                {"non-negative": True},  # type: ignore[list-item]
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
        }
    )
    wanted_parameters = Parameters.from_dict(
        {
            "kinetic": [
                ["1", 501e-3],
                ["2", 202e-4],
                ["3", 105e-5],
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
        }
    )

    time = np.arange(-10, 100, 1.5)
    pixel = np.arange(600, 750, 10)

    axis = {"time": time, "pixel": pixel}

    clp = create_gaussian_clp(["s1", "s2", "s3"], [7, 3, 30], [620, 670, 720], [10, 30, 50], pixel)


class ThreeComponentSequential:
    model = DecaySimpleModel(
        **{
            "megacomplex": {
                "mc1": {
                    "type": "decay-sequential",
                    "compartments": ["s1", "s2", "s3"],
                    "rates": [
                        "kinetic.1",
                        "kinetic.2",
                        "kinetic.3",
                    ],
                },
            },
            "irf": {
                "irf1": {
                    "type": "multi-gaussian",
                    "center": ["irf.center"],
                    "width": ["irf.width"],
                },
            },
            "dataset": {
                "dataset1": {
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = Parameters.from_dict(
        {
            "kinetic": [
                ["1", 501e-3],
                ["2", 202e-4],
                ["3", 105e-5],
                {"non-negative": True},  # type: ignore[list-item]
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
        }
    )
    wanted_parameters = Parameters.from_dict(
        {
            "kinetic": [
                ["1", 501e-3],
                ["2", 202e-4],
                ["3", 105e-5],
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
        }
    )

    time = np.arange(-10, 50, 1.0)
    pixel = np.arange(600, 750, 10)
    axis = {"time": time, "pixel": pixel}

    clp = create_gaussian_clp(["s1", "s2", "s3"], [7, 3, 30], [620, 670, 720], [10, 30, 50], pixel)


@pytest.mark.parametrize(
    "suite",
    [
        OneComponentOneChannel,
        OneComponentOneChannelGaussianIrf,
        ThreeComponentParallel,
        ThreeComponentSequential,
    ],
)
@pytest.mark.parametrize("nnls", [True, False])
def test_kinetic_model(suite, nnls):
    model = suite.model
    print(model.validate())
    assert model.valid()
    model.dataset_groups["default"].method = (
        "non_negative_least_squares" if nnls else "variable_projection"
    )

    wanted_parameters = suite.wanted_parameters
    print(model.validate(wanted_parameters))
    print(wanted_parameters)
    assert model.valid(wanted_parameters)

    initial_parameters = suite.initial_parameters
    print(model.validate(initial_parameters))
    assert model.valid(initial_parameters)

    print(model.markdown(initial_parameters))

    dataset = simulate(model, "dataset1", wanted_parameters, suite.axis, suite.clp)

    assert dataset.data.shape == (suite.axis["time"].size, suite.axis["pixel"].size)

    data = {"dataset1": dataset}

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=20,
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for param in result.optimized_parameters.all():
        assert np.allclose(param.value, wanted_parameters.get(param.label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(dataset["time"], resultdata["time"])
    assert np.array_equal(dataset["pixel"], resultdata["pixel"])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, rtol=1e-2)

    if suite is OneComponentOneChannelGaussianIrf:
        assert "irf_shift" in resultdata
        expected_center = wanted_parameters.get("2").value
        expected_shift = wanted_parameters.get("4").value
        assert np.allclose(resultdata["irf_shift"].values, expected_center - expected_shift)

    if len(model.irf) != 0:
        assert "irf" in resultdata


def test_finalize_data():
    model = DecayModel(
        **{
            "initial_concentration": {
                "j1": {"compartments": ["s1", "s2"], "parameters": ["3", "3"]},
            },
            "megacomplex": {
                "mc1": {"type": "decay", "k_matrix": ["k1"]},
                "mc2": {"type": "decay", "k_matrix": ["k2"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                },
                "k2": {
                    "matrix": {
                        ("s2", "s2"): "2",
                    }
                },
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1", "mc2"],
                },
            },
        }
    )

    parameters = Parameters.from_list(
        [101e-4, 101e-3, [1, {"vary": False, "non-negative": False}]]
    )

    time = np.arange(0, 50, 1.5)
    pixel = np.asarray([0, 2])
    axis = {"time": time, "pixel": pixel}

    clp = xr.DataArray(
        [[1, 0], [0, 1]], coords=[("pixel", pixel.data), ("clp_label", ["s1", "s2"])]
    )
    dataset = simulate(model, "dataset1", parameters, axis, clp)

    scheme = Scheme(
        model=model,
        parameters=parameters,
        data={"dataset1": dataset},
        maximum_number_function_evaluations=1,
    )
    result = optimize(scheme)

    result_dataset = result.data["dataset1"]

    assert "initial_concentration" in result_dataset
    assert "species_concentration" in result_dataset
    assert "species_associated_images" in result_dataset

    assert "decay_associated_images_mc1" in result_dataset
    assert "decay_associated_images_mc2" in result_dataset

    assert "k_matrix_mc1" in result_dataset
    assert "k_matrix_mc2" in result_dataset
    assert "k_matrix_reduced_mc1" in result_dataset
    assert "k_matrix_reduced_mc2" in result_dataset
    assert "a_matrix_mc1" in result_dataset
    assert "a_matrix_mc2" in result_dataset

    assert "species_mc1" in result_dataset.coords
    assert "species_mc2" in result_dataset.coords
    assert "initial_concentration_mc1" in result_dataset.coords
    assert "initial_concentration_mc2" in result_dataset.coords
    assert "component_mc1" in result_dataset.coords
    assert "component_mc2" in result_dataset.coords
    assert "rate_mc1" in result_dataset.coords
    assert "rate_mc2" in result_dataset.coords
    assert "lifetime_mc1" in result_dataset.coords
    assert "lifetime_mc2" in result_dataset.coords



"""Module containing the estimation provider classes."""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.model import DatasetGroup
from glotaran.model import DatasetModel
from glotaran.model import EqualAreaPenalty
from glotaran.model.dataset_model import has_dataset_model_global_model
from glotaran.model.item import fill_item
from glotaran.optimization.data_provider import DataProvider
from glotaran.optimization.data_provider import DataProviderLinked
from glotaran.optimization.matrix_provider import MatrixProviderLinked
from glotaran.optimization.matrix_provider import MatrixProviderUnlinked
from glotaran.optimization.nnls import residual_nnls
from glotaran.optimization.variable_projection import residual_variable_projection

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike
SUPPORTED_RESIUDAL_FUNCTIONS = {
    "variable_projection": residual_variable_projection,
    "non_negative_least_squares": residual_nnls,
}


class UnsupportedResidualFunctionError(ValueError):
    """Inidcates that the residual function is unsupported."""

    def __init__(self, residual_function: str):
        """Initialize an UnsupportedMethodError.

        Parameters
        ----------
        residual_function : str
            The unsupported residual_function.
        """
        super().__init__(
            f"Unknown residual function '{residual_function}', "
            f"supported functions are: {list(SUPPORTED_RESIUDAL_FUNCTIONS.keys())}."
        )


class EstimationProvider:
    """A class to provide estimation for optimization."""

    def __init__(self, dataset_group: DatasetGroup):
        """Initialize an estimation provider for a dataset group.

        Parameters
        ----------
        dataset_group : DatasetGroup
            The dataset group.

        Raises
        ------
        UnsupportedResidualFunctionError
            Raised when residual function of the group dataset group is unsupported.
        """
        self._group = dataset_group
        self._clp_penalty: list[float] = []
        try:
            self._residual_function = SUPPORTED_RESIUDAL_FUNCTIONS[dataset_group.residual_function]
        except KeyError as e:
            raise UnsupportedResidualFunctionError(dataset_group.residual_function) from e

    @property
    def group(self) -> DatasetGroup:
        """Get the dataset group.

        Returns
        -------
        DatasetGroup
            The dataset group.
        """
        return self._group

    def calculate_residual(
        self, matrix: ArrayLike, data: ArrayLike
    ) -> tuple[ArrayLike, ArrayLike]:
        """Calculate the clps and the residual for a matrix and data.

        Parameters
        ----------
        matrix : ArrayLike
            The matrix.
        data : ArrayLike
            The data.

        Returns
        -------
        tuple[ArrayLike, ArrayLike]
            The estimated clp and residual.
        """
        return self._residual_function(matrix, data)

    def retrieve_clps(
        self,
        clp_labels: list[str],
        reduced_clp_labels: list[str],
        reduced_clps: ArrayLike,
        index: int,
    ) -> ArrayLike:
        """Retrieve clp from reduced clp.

        Parameters
        ----------
        clp_labels : list[str]
            The original clp labels.
        reduced_clp_labels : list[str]
            The reduced clp labels.
        reduced_clps : ArrayLike
            The reduced clps.
        index : int
            The index on the global axis.

        Returns
        -------
        ArrayLike
            The retrieved clps.
        """
        model = self.group.model
        parameters = self.group.parameters
        if len(model.clp_relations) == 0 and len(model.clp_constraints) == 0:
            return reduced_clps

        clps = np.zeros(len(clp_labels))

        for i, label in enumerate(reduced_clp_labels):
            idx = clp_labels.index(label)
            clps[idx] = reduced_clps[i]

        for relation in model.clp_relations:
            relation = fill_item(relation, model, parameters)  # type:ignore[arg-type]
            if (
                relation.target in clp_labels
                and relation.applies(index)
                and relation.source in clp_labels
            ):
                source_idx = clp_labels.index(relation.source)
                target_idx = clp_labels.index(relation.target)
                clps[target_idx] = relation.parameter * clps[source_idx]
        return clps

    def get_additional_penalties(self) -> list[float]:
        """Get the additional penalty.

        Returns
        -------
        list[float]
            The additional penalty.
        """
        return self._clp_penalty

    def calculate_clp_penalties(
        self,
        clp_labels: list[list[str]],
        clps: list[np.ndarray],
        global_axis: np.ndarray,
    ) -> list[float]:
        """Calculate the clp penalty.

        Parameters
        ----------
        clp_labels : list[list[str]]
            The clp labels.
        clps : list[ArrayLike]
            The clps.
        global_axis : ArrayLike
            The global axis.

        Returns
        -------
        list[float]
            The clp penalty.
        """
        model = self.group.model
        parameters = self.group.parameters
        penalties = []
        for penalty in model.clp_penalties:
            if not isinstance(penalty, EqualAreaPenalty):
                continue
            penalty = fill_item(penalty, model, parameters)  # type:ignore[arg-type]

            source_area = _get_area(
                penalty.source,
                clp_labels,
                clps,
                penalty.source_intervals,
                global_axis,
            )
            target_area = _get_area(
                penalty.target,
                clp_labels,
                clps,
                penalty.target_intervals,
                global_axis,
            )

            if len(target_area) == 0 and len(source_area) == 0:
                continue
            elif len(target_area) == 0:
                warnings.warn(
                    "Ignoring equal area penalty, target clp " f"{penalty.target} not present."
                )
                continue
            elif len(source_area) == 0:
                warnings.warn(
                    "Ignoring equal area penalty, source clp " f"{penalty.source} not present."
                )
                continue

            area_penalty = np.abs(np.sum(source_area) - penalty.parameter * np.sum(target_area))

            penalties.append(area_penalty * penalty.weight)

        return penalties

    def estimate(self):
        """Calculate the estimation.

        .. # noqa: DAR401
        """
        raise NotImplementedError

    def get_full_penalty(self) -> ArrayLike:
        """Get the full penalty.

        Returns
        -------
        ArrayLike
            The clp penalty.

        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError

    def get_result(
        self,
    ) -> tuple[dict[str, xr.DataArray], dict[str, xr.DataArray],]:
        """Get the results of the estimation.

        Returns
        -------
        tuple[dict[str, xr.DataArray], dict[str, xr.DataArray]]
            A tuple of the estimated clps and residuals.

        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError


class EstimationProviderUnlinked(EstimationProvider):
    """A class to provide estimation for optimization of an unlinked dataset group."""

    def __init__(
        self,
        dataset_group: DatasetGroup,
        data_provider: DataProvider,
        matrix_provider: MatrixProviderUnlinked,
    ):
        """Initialize an estimation provider for an unlinked dataset group.

        Parameters
        ----------
        dataset_group : DatasetGroup
            The dataset group.
        data_provider : DataProvider
            The data provider.
        matrix_provider : MatrixProviderUnlinked
            The matrix provider.
        """
        super().__init__(dataset_group)
        self._data_provider = data_provider
        self._matrix_provider = matrix_provider
        self._clps: dict[str, list[ArrayLike] | ArrayLike] = {
            label: [] for label in self.group.dataset_models
        }
        self._residuals: dict[str, list[ArrayLike] | ArrayLike] = {
            label: [] for label in self.group.dataset_models
        }

    def estimate(self):
        """Calculate the estimation."""
        self._clp_penalty.clear()

        for dataset_model in self.group.dataset_models.values():
            if has_dataset_model_global_model(dataset_model):
                self.calculate_full_model_estimation(dataset_model)
            else:
                self.calculate_estimation(dataset_model)

    def get_full_penalty(self) -> ArrayLike:
        """Get the full penalty.

        Returns
        -------
        ArrayLike
            The clp penalty.
        """
        full_penalty = np.concatenate(
            [
                self._residuals[label]
                if has_dataset_model_global_model(dataset_model)
                else np.concatenate(self._residuals[label])
                for label, dataset_model in self.group.dataset_models.items()
            ]
        )
        if len(self._clp_penalty) != 0:
            full_penalty = np.concatenate([full_penalty, self._clp_penalty])
        return full_penalty

    def get_result(
        self,
    ) -> tuple[dict[str, list[xr.DataArray]], dict[str, list[xr.DataArray]],]:
        """Get the results of the estimation.

        Returns
        -------
        tuple[dict[str, xr.DataArray], dict[str, xr.DataArray]]
            A tuple of the estimated clps and residuals.
        """
        clps, residuals = {}, {}
        for label, dataset_model in self.group.dataset_models.items():
            model_dimension = self._data_provider.get_model_dimension(label)
            model_axis = self._data_provider.get_model_axis(label)
            global_dimension = self._data_provider.get_global_dimension(label)
            global_axis = self._data_provider.get_global_axis(label)

            if has_dataset_model_global_model(dataset_model):
                residuals[label] = xr.DataArray(
                    np.array(self._residuals[label]).T.reshape(model_axis.size, global_axis.size),
                    coords={global_dimension: global_axis, model_dimension: model_axis},
                    dims=[model_dimension, global_dimension],
                )
                clp_labels = self._matrix_provider.get_matrix_container(label).clp_labels
                global_clp_labels = self._matrix_provider.get_global_matrix_container(
                    label
                ).clp_labels
                clps[label] = xr.DataArray(
                    np.array(self._clps[label]).reshape((len(global_clp_labels), len(clp_labels))),
                    coords={"global_clp_label": global_clp_labels, "clp_label": clp_labels},
                    dims=["global_clp_label", "clp_label"],
                )

            else:
                residuals[label] = xr.DataArray(
                    np.array(self._residuals[label]).T,
                    coords={global_dimension: global_axis, model_dimension: model_axis},
                    dims=[model_dimension, global_dimension],
                )
                clps[label] = xr.DataArray(
                    self._clps[label],
                    coords=(
                        (global_dimension, global_axis),
                        (
                            "clp_label",
                            self._matrix_provider.get_matrix_container(label).clp_labels,
                        ),
                    ),
                )
        return clps, residuals

    def calculate_full_model_estimation(self, dataset_model: DatasetModel):
        """Calculate the estimation for a dataset with a full model.

        Parameters
        ----------
        dataset_model : DatasetModel
            The dataset model.
        """
        label = dataset_model.label
        full_matrix = self._matrix_provider.get_full_matrix(label)
        data = self._data_provider.get_flattened_data(label)
        self._clps[label], self._residuals[label] = self.calculate_residual(full_matrix, data)

    def calculate_estimation(self, dataset_model: DatasetModel):
        """Calculate the estimation for a dataset.

        Parameters
        ----------
        dataset_model : DatasetModel
            The dataset model.
        """
        label = dataset_model.label
        self._clps[label].clear()  # type:ignore[union-attr]
        self._residuals[label].clear()  # type:ignore[union-attr]

        global_axis = self._data_provider.get_global_axis(label)
        data = self._data_provider.get_data(label)
        clp_labels = []

        for index, global_index_value in enumerate(global_axis):
            matrix_container = self._matrix_provider.get_prepared_matrix_container(label, index)
            reduced_clps, residual = self.calculate_residual(
                matrix_container.matrix, data[:, index]
            )
            clp_labels.append(self._matrix_provider.get_matrix_container(label).clp_labels)
            clp = self.retrieve_clps(
                clp_labels[index], matrix_container.clp_labels, reduced_clps, global_index_value
            )

            self._clps[label].append(clp)  # type:ignore[union-attr]
            self._residuals[label].append(residual)  # type:ignore[union-attr]

        self._clp_penalty += self.calculate_clp_penalties(
            clp_labels, self._clps[label], global_axis  # type:ignore[arg-type]
        )


class EstimationProviderLinked(EstimationProvider):
    """A class to provide estimation for optimization of a linked dataset group."""

    def __init__(
        self,
        dataset_group: DatasetGroup,
        data_provider: DataProviderLinked,
        matrix_provider: MatrixProviderLinked,
    ):
        """Initialize an estimation provider for a linked dataset group.

        Parameters
        ----------
        dataset_group : DatasetGroup
            The dataset group.
        data_provider : DataProviderLinked
            The data provider.
        matrix_provider : MatrixProviderLinked
            The matrix provider.
        """
        super().__init__(dataset_group)
        self._data_provider = data_provider
        self._matrix_provider = matrix_provider
        self._clps: list[ArrayLike] = [
            None  # type:ignore[list-item]
        ] * self._data_provider.aligned_global_axis.size
        self._residuals: list[ArrayLike] = [
            None  # type:ignore[list-item]
        ] * self._data_provider.aligned_global_axis.size

    def estimate(self):
        """Calculate the estimation."""
        for index, global_index_value in enumerate(self._data_provider.aligned_global_axis):
            matrix_container = self._matrix_provider.get_aligned_matrix_container(index)
            data = self._data_provider.get_aligned_data(index)
            reduced_clps, residual = self.calculate_residual(matrix_container.matrix, data)
            self._clps[index] = self.retrieve_clps(
                self._matrix_provider.aligned_full_clp_labels[index],
                matrix_container.clp_labels,
                reduced_clps,
                global_index_value,
            )
            self._residuals[index] = residual

        self._clp_penalty = self.calculate_clp_penalties(
            self._matrix_provider.aligned_full_clp_labels,
            self._clps,
            self._data_provider.aligned_global_axis,
        )

    def get_full_penalty(self) -> ArrayLike:
        """Get the full penalty.

        Returns
        -------
        ArrayLike
            The clp penalty.
        """
        return np.concatenate((np.concatenate(self._residuals), self._clp_penalty))

    def get_result(
        self,
    ) -> tuple[dict[str, xr.DataArray], dict[str, xr.DataArray],]:
        """Get the results of the estimation.

        Returns
        -------
        tuple[dict[str, xr.DataArray], dict[str, xr.DataArray]]
            A tuple of the estimated clps and residuals.
        """
        clps: dict[str, xr.DataArray] = {}
        residuals: dict[str, xr.DataArray] = {}
        for dataset_label in self.group.dataset_models:
            dataset_clps, dataset_residual = [], []
            for index in range(self._data_provider.aligned_global_axis.size):
                group_label = self._data_provider.get_aligned_group_label(index)
                if dataset_label not in group_label:
                    continue

                group_datasets = self._data_provider.group_definitions[group_label]
                dataset_index = group_datasets.index(dataset_label)

                clp_labels = self._matrix_provider.get_matrix_container(dataset_label).clp_labels

                dataset_clps.append(
                    xr.DataArray(
                        [
                            self._clps[index][
                                self._matrix_provider.aligned_full_clp_labels[index].index(label)
                            ]
                            for label in clp_labels
                        ],
                        coords={"clp_label": clp_labels},
                    )
                )

                start = sum(
                    self._data_provider.get_model_axis(label).size
                    for label in group_datasets[:dataset_index]
                )
                end = start + self._data_provider.get_model_axis(dataset_label).size
                dataset_residual.append(self._residuals[index][start:end])

            model_dimension = self._data_provider.get_model_dimension(dataset_label)
            model_axis = self._data_provider.get_model_axis(dataset_label)
            global_dimension = self._data_provider.get_global_dimension(dataset_label)
            global_axis = self._data_provider.get_global_axis(dataset_label)
            clps[dataset_label] = xr.concat(
                dataset_clps,
                dim=global_dimension,
            )
            clps[dataset_label].coords[global_dimension] = global_axis
            residuals[dataset_label] = xr.DataArray(
                np.array(dataset_residual).T,
                coords={global_dimension: global_axis, model_dimension: model_axis},
                dims=[model_dimension, global_dimension],
            )
        return clps, residuals


def _get_area(
    clp_label: str,
    clp_labels: list[str] | list[list[str]],
    clps: list[ArrayLike],
    intervals: list[tuple[float, float]],
    global_axis: ArrayLike,
) -> ArrayLike:
    """Get get slice of a clp on intervals on the global axis.

    Parameters
    ----------
    clp_label : str
        The label of the clp.
    clp_labels: list[str] | list[list[str]]
        The clp labels.
    clps : list[ArrayLike]
        The clps.
    intervals: list[tuple[float, float]]
        The intervals on the global axis.
    global_axis : ArrayLike
        The global axis.

    Returns
    -------
    ArrayLike:
        The concatenated slices.
    """
    area = []

    for interval in intervals:
        if interval[0] > global_axis[-1]:
            continue
        bounded_interval = (
            max(interval[0], np.min(global_axis)),
            min(interval[1], np.max(global_axis)),
        )

        interval_slice = DataProvider.get_axis_slice_from_interval(bounded_interval, global_axis)
        start_idx, end_idx = interval_slice.start, interval_slice.stop

        for i in range(start_idx, end_idx):
            index_clp_labels: list[str] = (
                clp_labels[i]
                if isinstance(clp_labels[0], list)
                else clp_labels  # type:ignore[assignment]
            )
            if clp_label in index_clp_labels:
                area.append(clps[i][index_clp_labels.index(clp_label)])

    return np.asarray(area)  # TODO: normalize for distance on global axis



import click

from glotaran.cli.commands import util


@util.signature_analysis
def print_cmd(parameters_file: str, model_file: str, scheme_file: str):
    """Parses scheme, a model or a parameter file and prints the result as a Markdown formatted
    string."""
    model = None
    parameters = None
    if scheme_file is not None:
        scheme = util.load_scheme_file(scheme_file, verbose=False)
        click.echo(scheme.markdown())
        return
    if model_file is not None:
        model = util.load_model_file(model_file, verbose=False)

    if parameters_file is not None:
        parameters = util.load_parameters_file(parameters_file, verbose=False)

    if model:
        click.echo(model.markdown(parameters=parameters))
    elif parameters:
        click.echo(parameters.markdown())
    else:
        click.echo("Nothing to print, please type 'glotaran print --help' for more info.")



"""Functions for simulating a dataset using a global optimization model."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.model import DatasetModel
from glotaran.model.dataset_model import get_dataset_model_model_dimension
from glotaran.model.dataset_model import has_dataset_model_global_model
from glotaran.model.item import fill_item
from glotaran.optimization.matrix_provider import MatrixProvider

if TYPE_CHECKING:
    from glotaran.model import Model
    from glotaran.parameter import Parameters
    from glotaran.typing.types import ArrayLike


def simulate(
    model: Model,
    dataset: str,
    parameters: Parameters,
    coordinates: dict[str, ArrayLike],
    clp: xr.DataArray | None = None,
    noise: bool = False,
    noise_std_dev: float = 1.0,
    noise_seed: int | None = None,
) -> xr.Dataset:
    """Simulate a dataset using a model.

    Parameters
    ----------
    model : Model
        The model containing the dataset model.
    dataset : str
        Label of the dataset to simulate
    parameters : Parameters
        The parameters for the simulation.
    coordinates : dict[str, ArrayLike]
        A dictionary with the coordinates used for simulation (e.g. time, wavelengths, ...).
    clp : xr.DataArray | None
        A matrix with conditionally linear parameters (e.g. spectra, pixel intensity, ...).
        Will be used instead of the dataset's global megacomplexes if not None.
    noise : bool
        Add noise to the simulation.
    noise_std_dev : float
        The standard deviation for noise simulation.
    noise_seed : int | None
        The seed for the noise simulation.

    Returns
    -------
    xr.Dataset
        The simulated dataset.


    Raises
    ------
    ValueError
        Raised if dataset model has no global megacomplex and no clp are provided.
    """
    dataset_model = fill_item(model.dataset[dataset], model, parameters)
    model_dimension = get_dataset_model_model_dimension(dataset_model)
    model_axis = coordinates[model_dimension]
    global_dimension = next(dim for dim in coordinates if dim != model_dimension)
    global_axis = coordinates[global_dimension]

    if has_dataset_model_global_model(dataset_model):
        result = simulate_full_model(
            dataset_model, global_dimension, global_axis, model_dimension, model_axis
        )
    elif clp is None:
        raise ValueError(
            f"Cannot simulate dataset '{dataset}'. "
            "No global megacomplex is defined and no clp provided."
        )
    else:
        result = simulate_from_clp(
            dataset_model, global_dimension, global_axis, model_dimension, model_axis, clp
        )

    if noise:
        if noise_seed is not None:
            np.random.seed(noise_seed)
        result["data"] = (result.data.dims, np.random.normal(result.data, noise_std_dev))

    return result


def simulate_from_clp(
    dataset_model: DatasetModel,
    global_dimension: str,
    global_axis: ArrayLike,
    model_dimension: str,
    model_axis: ArrayLike,
    clp: xr.DataArray,
) -> xr.Dataset:
    """Simulate a dataset model from pre-defined conditionally linear parameters.

    Parameters
    ----------
    dataset_model : DatasetModel
        The dataset model to simulate.
    global_dimension : str
        The global dimension of the dataset.
    global_axis : ArrayLike
        The global axis of the dataset.
    model_dimension : str
        The model dimension of the dataset.
    model_axis : ArrayLike
        The model axis of the dataset.
    clp : xr.DataArray
        A matrix with conditionally linear parameters.

    Returns
    -------
    xr.Dataset
        The simulated dataset.

    Raises
    ------
    ValueError
        Raised if the clp are missing the dimension 'clp_label'.
    """
    if "clp_label" not in clp.coords:
        raise ValueError("Missing coordinate 'clp_label' in clp.")

    matrix = MatrixProvider.calculate_dataset_matrix(dataset_model, global_axis, model_axis)
    result = xr.DataArray(
        np.zeros((model_axis.size, global_axis.size)),
        coords=[
            (model_dimension, model_axis),
            (global_dimension, global_axis),
        ],
    )
    result = result.to_dataset(name="data")
    for i in range(global_axis.size):
        this_matrix = matrix.matrix[i] if matrix.is_index_dependent else matrix.matrix
        result.data[:, i] = np.dot(
            this_matrix,
            clp.isel({global_dimension: i}).sel({"clp_label": matrix.clp_labels}),
        )

    return result


def simulate_full_model(
    dataset_model: DatasetModel,
    global_dimension: str,
    global_axis: ArrayLike,
    model_dimension: str,
    model_axis: ArrayLike,
) -> xr.Dataset:
    """Simulate a dataset model with global megacomplexes.

    Parameters
    ----------
    dataset_model : DatasetModel
        The dataset model to simulate.
    global_dimension : str
        The global dimension of the dataset.
    global_axis : ArrayLike
        The global axis of the dataset.
    model_dimension : str
        The model dimension of the dataset.
    model_axis : ArrayLike
        The model axis of the dataset.

    Returns
    -------
    xr.Dataset
        The simulated dataset.

    Raises
    ------
    ValueError
        Raised if at least one of the dataset model's global megacomplexes is index dependent.
    """
    global_matrix = MatrixProvider.calculate_dataset_matrix(
        dataset_model, global_axis, model_axis, global_matrix=True
    )
    if global_matrix.is_index_dependent:
        raise ValueError("Index dependent models for global dimension are not supported.")
    global_clp_labels = global_matrix.clp_labels
    global_matrix = xr.DataArray(
        global_matrix.matrix.T,
        coords=[
            ("clp_label", global_clp_labels),
            (global_dimension, global_axis),
        ],
    )

    return simulate_from_clp(
        dataset_model, global_dimension, global_axis, model_dimension, model_axis, global_matrix
    )



"""Module containing TSV io support."""

from __future__ import annotations

from typing import TYPE_CHECKING

from glotaran.io import ProjectIoInterface
from glotaran.io import load_parameters
from glotaran.io import register_project_io
from glotaran.io import save_parameters

if TYPE_CHECKING:
    from glotaran.parameter import Parameters


@register_project_io(["tsv"])
class TsvProjectIo(ProjectIoInterface):
    """Plugin for TSV data io."""

    def load_parameters(self, file_name: str) -> Parameters:
        """Load parameters from TSV file.

        Parameters
        ----------
        file_name : str
            Name of file to be loaded.

        Returns
        -------
            :class:`Parameters`
        """
        return load_parameters(file_name, format_name="csv", sep="\t")

    def save_parameters(
        self,
        parameters: Parameters,
        file_name: str,
        *,
        replace_infinfinity: bool = True,
    ) -> None:
        """Save a :class:`Parameters` to a TSV file.

        Parameters
        ----------
        parameters : Parameters
            Parameters to be saved to file.
        file_name : str
            File to write the parameters to.
        replace_infinfinity : bool
            Weather to replace infinity values with empty strings.
        """
        save_parameters(
            parameters,
            file_name,
            format_name="csv",
            sep="\t",
            replace_infinfinity=replace_infinfinity,
        )



"""The parameters class."""
from __future__ import annotations

from collections.abc import Generator
from textwrap import indent
from typing import TYPE_CHECKING
from typing import Any

import asteval
import numpy as np
import pandas as pd
from tabulate import tabulate

from glotaran.io import load_parameters
from glotaran.parameter.parameter import Parameter
from glotaran.utils.ipython import MarkdownStr
from glotaran.utils.sanitize import pretty_format_numerical

if TYPE_CHECKING:
    from glotaran.parameter.parameter_history import ParameterHistory


class ParameterNotFoundException(Exception):
    """Raised when a Parameter is not found."""

    def __init__(self, label: str):  # noqa: D107
        super().__init__(f"Cannot find parameter {label}")


class Parameters:
    """A container for :class:`Parameter`."""

    loader = load_parameters

    def __init__(self, parameters: dict[str, Parameter]):
        """Create :class:`Parameters`.

        Parameters
        ----------
        parameters : dict[str, Parameter]
            A parameter list containing parameters

        Returns
        -------
        'Parameters'
            The created :class:`Parameters`.
        """
        self._parameters: dict[str, Parameter] = parameters
        self._evaluator = asteval.Interpreter(symtable=asteval.make_symbol_table(parameters=self))
        self.source_path = "parameters.csv"
        self.update_parameter_expression()

    @classmethod
    def from_list(
        cls, parameter_list: list[float | int | str | list[Any] | dict[str, Any]]
    ) -> Parameters:
        """Create :class:`Parameters` from a list.

        Parameters
        ----------
        parameter_list : list[float | list[Any]]
            A parameter list containing parameters

        Returns
        -------
        Parameters
            The created :class:`Parameters`.

        .. # noqa: D414
        """
        defaults: dict[str, Any] | None = next(
            (item for item in parameter_list if isinstance(item, dict)), None
        )
        parameters = {}

        for i, item in enumerate(item for item in parameter_list if not isinstance(item, dict)):
            if not isinstance(item, list):
                item = [item]
            if not any(isinstance(v, str) for v in item):
                item += [f"{i+1}"]
            parameter = Parameter.from_list(item, default_options=defaults)
            parameters[parameter.label] = parameter
        return cls(parameters)

    @classmethod
    def from_dict(
        cls,
        parameter_dict: dict[str, dict[str, Any] | list[float | list[Any]]],
    ) -> Parameters:
        """Create a :class:`Parameters` from a dictionary.

        Parameters
        ----------
        parameter_dict: dict[str, dict[str, Any] | list[float | list[Any]]]
            A parameter dictionary containing parameters.

        Returns
        -------
        Parameters
            The created :class:`Parameters`

        .. # noqa: D414
        """
        parameters = {}
        for label, param_def, default in flatten_parameter_dict(parameter_dict):
            parameter = Parameter.from_list(param_def, default_options=default)
            label += f".{parameter.label}"
            parameter.label = label
            parameters[label] = parameter

        return cls(parameters)

    @classmethod
    def from_parameter_dict_list(cls, parameter_dict_list: list[dict[str, Any]]) -> Parameters:
        """Create :class:`Parameters` from a list of parameter dictionaries.

        Parameters
        ----------
        parameter_dict_list : list[dict[str, Any]]
            A list of parameter dictionaries.

        Returns
        -------
        Parameters
            The created :class:`Parameters`.

        .. # noqa: D414
        """
        parameters = {}
        for parameter_dict in parameter_dict_list:
            parameter = Parameter(**parameter_dict)
            parameters[parameter.label] = parameter
        return cls(parameters)

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, source: str = "DataFrame") -> Parameters:
        """Create a :class:`Parameters` from a :class:`pandas.DataFrame`.

        Parameters
        ----------
        df : pd.DataFrame
            The source data frame.
        source : str
            Optional name of the source file, used for error messages.

        Raises
        ------
        ValueError
            Raised if the columns 'label' or 'value' doesn't exist. Also raised if the columns
            'minimum', 'maximum' or 'values' contain non numeric values or if the columns
            'non-negative' or 'vary' are no boolean.

        Returns
        -------
        Parameters
            The created parameter group.

        .. # noqa: D414
        """
        for column_name in ["label", "value"]:
            if column_name not in df:
                raise ValueError(f"Missing required column '{column_name}' in '{source}'.")

        for column_name in filter(lambda x: x in df.columns, ["minimum", "maximum", "value"]):
            if any(not np.isreal(v) for v in df[column_name]):
                raise ValueError(f"Column '{column_name}' in '{source}' has non numeric values.")

        for column_name in filter(lambda x: x in df.columns, ["non_negative", "vary"]):
            df[column_name] = [v != 0 if isinstance(v, int) else v for v in df[column_name]]
            if any(not isinstance(v, bool) for v in df[column_name]):
                raise ValueError(f"Column '{column_name}' in '{source}' has non boolean values.")

        # clean NaN if expressions
        if "expression" in df:
            expressions = df["expression"].to_list()
            df["expression"] = [expr if isinstance(expr, str) else None for expr in expressions]
        return cls.from_parameter_dict_list(df.to_dict(orient="records"))

    def to_dataframe(self) -> pd.DataFrame:
        """Create a pandas data frame from the group.

        Returns
        -------
        pd.DataFrame
            The created data frame.
        """
        return pd.DataFrame(self.to_parameter_dict_list())

    def to_parameter_dict_list(self) -> list[dict[str, Any]]:
        """Create list of parameter dictionaries from the group.

        Returns
        -------
        list[dict[str, Any]]
            A list of parameter dictionaries.
        """
        return [p.as_dict() for p in self.all()]

    def to_parameter_dict_or_list(self, serialize_parameters: bool = False) -> dict | list:
        """Convert to a dict or list of parameter definitions.

        Parameters
        ----------
        serialize_parameters : bool
            If true, the parameters will be serialized into list representation.

        Returns
        -------
        dict | list
            A dict or list of parameter definitions.
        """
        if all("." not in p.label for p in self.all()):
            return list(self.all())
        parameter_dict: dict[str, Any] = {}
        for parameter in self.all():
            path = parameter.label.split(".")
            nodes = path[:-2]
            node = parameter_dict
            for n in nodes:
                if n not in node:
                    node[n] = {}
                node = node[n]
            upper_node = path[-2]
            if upper_node not in node:
                node[upper_node] = []
            node[upper_node].append(
                parameter.as_list(label_short=True) if serialize_parameters else parameter
            )
        return parameter_dict

    def set_from_history(self, history: ParameterHistory, index: int):
        """Update the :class:`Parameters` with values from a parameter history.

        Parameters
        ----------
        history : ParameterHistory
            The parameter history.
        index : int
            The history index.
        """
        self.set_from_label_and_value_arrays(
            # Omit 0th element with `iteration` label
            history.parameter_labels[1:],
            history.get_parameters(index)[1:],
        )

    def copy(self) -> Parameters:
        """Create a copy of the :class:`Parameters`.

        Returns
        -------
        Parameters :
            A copy of the :class:`Parameters`.

        .. # noqa: D414
        """
        return Parameters(
            {label: parameter.copy() for label, parameter in self._parameters.items()}
        )

    def all(self) -> Generator[Parameter, None, None]:
        """Iterate over all parameters.

        Yields
        ------
        Parameter
            A parameter in the parameters.
        """
        yield from self._parameters.values()

    def has(self, label: str) -> bool:
        """Check if a parameter with the given label is in the group or in a subgroup.

        Parameters
        ----------
        label : str
            The label of the parameter, with its path in a :class:`ParameterGroup` prepended.

        Returns
        -------
        bool
            Whether a parameter with the given label exists in the group.
        """
        return label in self._parameters

    def get(self, label: str) -> Parameter:
        """Get a :class:`Parameter` by its label.

        Parameters
        ----------
        label : str
            The label of the parameter, with its path in a :class:`ParameterGroup` prepended.

        Returns
        -------
        Parameter
            The parameter.

        Raises
        ------
        ParameterNotFoundException
            Raised if no parameter with the given label exists.
        """
        try:
            return self._parameters[label]
        except KeyError as error:
            raise ParameterNotFoundException(label) from error

    def update_parameter_expression(self):
        """Update all parameters which have an expression.

        Raises
        ------
        ValueError
            Raised if an expression evaluates to a non-numeric value.
        """
        for parameter in self.all():
            if parameter.expression is not None:
                value = self._evaluator(parameter.transformed_expression)
                if not isinstance(value, (int, float)):
                    raise ValueError(
                        f"Expression '{parameter.expression}' of parameter '{parameter.label}' "
                        f"evaluates to non numeric value '{value}'."
                    )
                parameter.value = value

    def get_label_value_and_bounds_arrays(
        self, exclude_non_vary: bool = False
    ) -> tuple[list[str], np.ndarray, np.ndarray, np.ndarray]:
        """Return a arrays of all parameter labels, values and bounds.

        Parameters
        ----------
        exclude_non_vary: bool
            If true, parameters with `vary=False` are excluded.

        Returns
        -------
        tuple[list[str], np.ndarray, np.ndarray, np.ndarray]
            A tuple containing a list of parameter labels and
            an array of the values, lower and upper bounds.
        """
        self.update_parameter_expression()

        labels = []
        values = []
        lower_bounds = []
        upper_bounds = []

        for parameter in self.all():
            if not exclude_non_vary or parameter.vary:
                labels.append(parameter.label)
                value, minimum, maximum = parameter.get_value_and_bounds_for_optimization()
                values.append(value)
                lower_bounds.append(minimum)
                upper_bounds.append(maximum)

        return labels, np.asarray(values), np.asarray(lower_bounds), np.asarray(upper_bounds)

    def set_from_label_and_value_arrays(self, labels: list[str], values: np.ndarray):
        """Update the parameter values from a list of labels and values.

        Parameters
        ----------
        labels : list[str]
            A list of parameter labels.
        values : np.ndarray
            An array of parameter values.

        Raises
        ------
        ValueError
            Raised if the size of the labels does not match the stize of values.
        """
        if len(labels) != len(values):
            raise ValueError(
                f"Length of labels({len(labels)}) not equal to length of values({len(values)})."
            )

        for label, value in zip(labels, values):
            self.get(label).set_value_from_optimization(value)

        self.update_parameter_expression()

    def markdown(self, float_format: str = ".3e") -> MarkdownStr:
        """Format the :class:`ParameterGroup` as markdown string.

        This is done by recursing the nested :class:`ParameterGroup` tree.

        Parameters
        ----------
        float_format: str
            Format string for floating point numbers, by default ".3e"

        Returns
        -------
        MarkdownStr :
            The markdown representation as string.
        """
        return param_dict_to_markdown(self.to_parameter_dict_or_list(), float_format=float_format)

    def _repr_markdown_(self) -> str:
        """Create a markdown representation.

        Special method used by ``ipython`` to render markdown.

        Returns
        -------
        str :
            The markdown representation as string.
        """
        return str(self.markdown())

    @property
    def labels(self) -> list[str]:
        """List of all labels.

        Returns
        -------
        list[str]
        """
        return sorted(p.label for p in self.all())

    def __str__(self) -> str:
        """Representation used by print and str."""
        return str(self.markdown())

    def __repr__(self) -> str:
        """Representation debug."""
        params = [f"{p.label!r}: {repr(p)}" for p in self.all()]
        return f"Parameters({{{', '.join(params)}}})"

    def __eq__(self, other: object) -> bool:
        """=="""  # noqa: D400
        if isinstance(other, Parameters):
            return self.labels == other.labels and all(
                self.get(label)._deep_equals(other.get(label)) for label in self.labels
            )
        raise NotImplementedError(
            "Parameters can only be compared with instances of Parameters, "
            f"not with {type(other).__qualname__!r}."
        )


def flatten_parameter_dict(
    parameter_dict: dict,
) -> Generator[tuple[str, list[Any], dict[str, Any] | None], None, None]:
    """Flatten a parameter dictionary.

    Parameters
    ----------
    parameter_dict: dict
        The parameter dictionary.

    Yields
    ------
    tuple[str, list[Any], dict | None
        The concatenated keys, the parameter definition and default options.
    """
    for key, value in parameter_dict.items():
        if isinstance(value, dict):
            for sub_key, sub_value, sub_dict in flatten_parameter_dict(value):
                yield f"{key}.{sub_key}", sub_value, sub_dict
        elif isinstance(value, list):
            sub_dict: dict[str, Any] | None = next(  # type:ignore[no-redef]
                (item for item in value if isinstance(item, dict)), None
            )
            for index, list_value in enumerate(
                (list_value for list_value in value if not isinstance(list_value, dict)), start=1
            ):
                if not isinstance(list_value, list):
                    list_value = [str(index), list_value]
                elif not any(isinstance(v, str) for v in list_value):
                    list_value += [str(index)]
                yield key, list_value, sub_dict


def param_dict_to_markdown(
    parameters: dict | list,
    float_format: str = ".3e",
    depth: int = 0,
    label: str | None = None,
) -> MarkdownStr:
    """Format the :class:`Parameters` as markdown string.

    This is done by recursing the nested :class:`Parameters` tree.

    Parameters
    ----------
    parameters: dict | list
        The parameter dict or list.
    float_format: str
        Format string for floating point numbers, by default ".3e"
    depth: int
        The depth of the parameter dict.
    label: str | None
        The label of the parameter dict.

    Returns
    -------
    MarkdownStr :
        The markdown representation as string.
    """
    node_indentation = "  " * depth
    return_string = ""
    table_header = [
        "_Label_",
        "_Value_",
        "_Standard Error_",
        "_t-value_",
        "_Minimum_",
        "_Maximum_",
        "_Vary_",
        "_Non-Negative_",
        "_Expression_",
    ]
    if label is not None:
        return_string += f"{node_indentation}* __{label}__:\n"
    if isinstance(parameters, list):
        parameter_rows = [
            [
                parameter.label_short,
                parameter.value,
                parameter.standard_error,
                repr(pretty_format_numerical(parameter.value / parameter.standard_error)),
                parameter.minimum,
                parameter.maximum,
                parameter.vary,
                parameter.non_negative,
                f"`{parameter.expression}`",
            ]
            for parameter in parameters
        ]
        parameter_table = indent(
            tabulate(
                parameter_rows,
                headers=table_header,
                tablefmt="github",
                missingval="None",
                floatfmt=float_format,
            ),
            f"  {node_indentation}",
        )
        return_string += f"\n{parameter_table}\n\n"
    else:
        for label, child in sorted(parameters.items()):
            return_string += str(
                param_dict_to_markdown(
                    child, float_format=float_format, depth=depth + 1, label=label
                )
            )
    return MarkdownStr(return_string.replace("'", " "))



from __future__ import annotations

import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

import glotaran
from glotaran.deprecation.deprecation_utils import GlotaranApiDeprecationWarning
from glotaran.deprecation.deprecation_utils import GlotaranDeprecatedApiError
from glotaran.deprecation.deprecation_utils import OverDueDeprecation
from glotaran.deprecation.deprecation_utils import check_overdue
from glotaran.deprecation.deprecation_utils import deprecate
from glotaran.deprecation.deprecation_utils import deprecate_dict_entry
from glotaran.deprecation.deprecation_utils import glotaran_version
from glotaran.deprecation.deprecation_utils import module_attribute
from glotaran.deprecation.deprecation_utils import parse_version
from glotaran.deprecation.deprecation_utils import raise_deprecation_error
from glotaran.deprecation.deprecation_utils import warn_deprecated

if TYPE_CHECKING:
    from collections.abc import Hashable
    from collections.abc import Mapping
    from typing import Any

    from _pytest.monkeypatch import MonkeyPatch
    from _pytest.recwarn import WarningsRecorder


DEP_UTILS_QUALNAME = "glotaran.deprecation.deprecation_utils"

DEPRECATION_QUAL_NAME = f"{DEP_UTILS_QUALNAME}.parse_version(version_str)"
NEW_QUAL_NAME = f"{DEP_UTILS_QUALNAME}.check_qualnames_in_tests(qualnames)"

DEPRECATION_WARN_MESSAGE = (
    "Usage of 'glotaran.deprecation.deprecation_utils.parse_version(version_str)' "
    "was deprecated, use "
    "'glotaran.deprecation.deprecation_utils.check_qualnames_in_tests(qualnames)' "
    "instead.\nThis usage will be an error in version: '0.6.0'."
)
DEPRECATION_ERROR_MESSAGE = (
    "Usage of 'glotaran.deprecation.deprecation_utils.parse_version(version_str)' was deprecated, "
    "use 'glotaran.deprecation.deprecation_utils.check_qualnames_in_tests(qualnames)' instead.\n"
    "It wasn't possible to restore the original behavior of this usage "
    "(most likely due to an object hierarchy change)."
    "This usage change message won't be show as of version: '0.6.0'."
)
OVERDUE_ERROR_MESSAGE = (
    "Support for 'glotaran.deprecation.deprecation_utils.parse_version' "
    "was supposed to be dropped in version: '0.6.0'.\n"
    "Current version is: '1.0.0'"
)


class DummyClass:
    """Dummy class to check check_qualnames_in_tests"""

    foo: dict[str, str] = {}


@pytest.fixture
def glotaran_0_3_0(monkeypatch: MonkeyPatch):
    """Mock glotaran version to always be 0.3.0 for the test."""
    monkeypatch.setattr(
        glotaran.deprecation.deprecation_utils, "glotaran_version", lambda: "0.3.0"
    )
    yield


@pytest.fixture
def glotaran_1_0_0(monkeypatch: MonkeyPatch):
    """Mock glotaran version to always be 1.0.0 for the test."""
    monkeypatch.setattr(
        glotaran.deprecation.deprecation_utils, "glotaran_version", lambda: "1.0.0"
    )
    yield


def test_glotaran_version():
    """Versions are the same."""
    assert glotaran_version() == glotaran.__version__


@pytest.mark.parametrize(
    "version_str, expected",
    (
        ("0.0.1", (0, 0, 1)),
        ("0.0.1.post", (0, 0, 1)),
        ("0.0.1-dev", (0, 0, 1)),
        ("0.0.1-dev.post", (0, 0, 1)),
    ),
)
def test_parse_version(version_str: str, expected: tuple[int, int, int]):
    """Valid version strings."""
    assert parse_version(version_str) == expected


@pytest.mark.parametrize(
    "version_str",
    ("1", "0.1", "a.b.c"),
)
def test_parse_version_errors(version_str: str):
    """Invalid version strings"""
    with pytest.raises(ValueError, match=f"'{version_str}'"):
        parse_version(version_str)


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_check_overdue_no_raise(monkeypatch: MonkeyPatch):
    """Current version smaller then drop_version."""
    check_overdue(
        deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
        to_be_removed_in_version="0.6.0",
    )


@pytest.mark.usefixtures("glotaran_1_0_0")
def test_check_overdue_raises(monkeypatch: MonkeyPatch):
    """Current version is equal or bigger than drop_version."""
    with pytest.raises(OverDueDeprecation) as excinfo:
        check_overdue(
            deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
            to_be_removed_in_version="0.6.0",
        )

    assert str(excinfo.value) == OVERDUE_ERROR_MESSAGE


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_raise_deprecation_error(monkeypatch: MonkeyPatch):
    """Current version smaller then drop_version."""
    with pytest.raises(GlotaranDeprecatedApiError) as excinfo:
        raise_deprecation_error(
            deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
            new_qual_name_usage=NEW_QUAL_NAME,
            to_be_removed_in_version="0.6.0",
        )
    assert str(excinfo.value) == DEPRECATION_ERROR_MESSAGE


@pytest.mark.usefixtures("glotaran_1_0_0")
def test_raise_deprecation_error_overdue(monkeypatch: MonkeyPatch):
    """Current version is equal or bigger than drop_version."""
    with pytest.raises(OverDueDeprecation) as excinfo:
        raise_deprecation_error(
            deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
            new_qual_name_usage=NEW_QUAL_NAME,
            to_be_removed_in_version="0.6.0",
        )
    assert str(excinfo.value) == OVERDUE_ERROR_MESSAGE


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_warn_deprecated():
    """Warning gets shown when all is in order."""
    with pytest.warns(GlotaranApiDeprecationWarning) as record:
        warn_deprecated(
            deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
            new_qual_name_usage=NEW_QUAL_NAME,
            to_be_removed_in_version="0.6.0",
        )

        assert len(record) == 1
        assert record[0].message.args[0] == DEPRECATION_WARN_MESSAGE
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.usefixtures("glotaran_1_0_0")
def test_warn_deprecated_overdue_deprecation(monkeypatch: MonkeyPatch):
    """Current version is equal or bigger than drop_version."""

    with pytest.raises(OverDueDeprecation) as excinfo:
        warn_deprecated(
            deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
            new_qual_name_usage=NEW_QUAL_NAME,
            to_be_removed_in_version="0.6.0",
        )
    assert str(excinfo.value) == OVERDUE_ERROR_MESSAGE


@pytest.mark.filterwarnings("ignore:Usage")
@pytest.mark.xfail(strict=True, reason="Dev version aren't checked")
def test_warn_deprecated_no_overdue_deprecation_on_dev(monkeypatch: MonkeyPatch):
    """Current version is equal or bigger than drop_version but it's a dev version."""
    monkeypatch.setattr(
        glotaran.deprecation.deprecation_utils, "glotaran_version", lambda: "0.6.0-dev"
    )

    with pytest.raises(OverDueDeprecation):
        warn_deprecated(
            deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
            new_qual_name_usage=NEW_QUAL_NAME,
            to_be_removed_in_version="0.6.0",
        )


@pytest.mark.parametrize(
    "deprecated_qual_name_usage,new_qual_name_usage",
    (
        (
            "glotaran.does_not_exists(foo)",
            DEPRECATION_QUAL_NAME,
        ),
        (
            DEPRECATION_QUAL_NAME,
            "glotaran.does_not_exists(foo)",
        ),
    ),
)
@pytest.mark.xfail(strict=True, reason="Should fail if any qualname is wrong.")
@pytest.mark.usefixtures("glotaran_0_3_0")
def test_warn_deprecated_broken_deprecated_qualname(
    deprecated_qual_name_usage: str, new_qual_name_usage: str
):
    """Fail if any qualname is wrong."""
    warn_deprecated(
        deprecated_qual_name_usage=deprecated_qual_name_usage,
        new_qual_name_usage=new_qual_name_usage,
        to_be_removed_in_version="0.6.0",
    )


@pytest.mark.parametrize(
    "deprecated_qual_name_usage,new_qual_name_usage,check_qualnames",
    (
        ("glotaran.does_not_exists(foo)", DEPRECATION_QUAL_NAME, (False, True)),
        (DEPRECATION_QUAL_NAME, "glotaran.does_not_exists(foo)", (True, False)),
    ),
)
@pytest.mark.usefixtures("glotaran_0_3_0")
def test_warn_deprecated_broken_qualname_no_check(
    deprecated_qual_name_usage: str, new_qual_name_usage: str, check_qualnames: tuple[bool, bool]
):
    """Not checking broken imports."""
    with pytest.warns(GlotaranApiDeprecationWarning):
        warn_deprecated(
            deprecated_qual_name_usage=deprecated_qual_name_usage,
            new_qual_name_usage=new_qual_name_usage,
            to_be_removed_in_version="0.6.0",
            check_qual_names=check_qualnames,
        )


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_warn_deprecated_sliced_method():
    """Slice away method for importing and check class for attribute"""
    with pytest.warns(GlotaranApiDeprecationWarning):
        warn_deprecated(
            deprecated_qual_name_usage=(
                "glotaran.deprecation.test.test_deprecation_utils.DummyClass.foo()"
            ),
            new_qual_name_usage=DEPRECATION_QUAL_NAME,
            to_be_removed_in_version="0.6.0",
            importable_indices=(2, 1),
        )


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_warn_deprecated_sliced_mapping():
    """Slice away mapping for importing and check class for attribute"""
    with pytest.warns(GlotaranApiDeprecationWarning):
        warn_deprecated(
            deprecated_qual_name_usage=(
                "glotaran.deprecation.test.test_deprecation_utils.DummyClass.foo['bar']"
            ),
            new_qual_name_usage=DEPRECATION_QUAL_NAME,
            to_be_removed_in_version="0.6.0",
            importable_indices=(2, 1),
        )


@pytest.mark.xfail(
    strict=True,
    reason="Should fail if new qualname is wrong, even if check_deprecated_qualname is False.",
)
@pytest.mark.usefixtures("glotaran_0_3_0")
def test_warn_deprecated_broken_new_qualname_no_check():
    """Fail if new qualname is wrong."""
    warn_deprecated(
        deprecated_qual_name_usage="glotaran.does_not_exists(foo)",
        new_qual_name_usage="glotaran.does_not_exists(foo)",
        to_be_removed_in_version="0.6.0",
        check_qual_names=(False, True),
    )


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecated_decorator_function(recwarn: WarningsRecorder):
    """Deprecate function with decorator."""
    warnings.simplefilter("always")

    @deprecate(
        deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
        new_qual_name_usage=NEW_QUAL_NAME,
        to_be_removed_in_version="0.6.0",
    )
    def dummy():
        """Dummy docstring for testing."""

    dummy()

    assert dummy.__doc__ == "Dummy docstring for testing."
    assert len(recwarn) == 1
    assert recwarn[0].category == GlotaranApiDeprecationWarning
    assert recwarn[0].message.args[0] == DEPRECATION_WARN_MESSAGE
    assert Path(recwarn[0].filename) == Path(__file__)

    dummy()

    assert len(recwarn) == 2


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecated_decorator_class(recwarn: WarningsRecorder):
    """Deprecate class with decorator."""
    warnings.simplefilter("always")

    @deprecate(
        deprecated_qual_name_usage=DEPRECATION_QUAL_NAME,
        new_qual_name_usage=NEW_QUAL_NAME,
        to_be_removed_in_version="0.6.0",
    )
    class Foo:
        """Foo class docstring for testing."""

        @classmethod
        def from_string(cls, string: str):
            """Just another method to init the class for testing."""
            return cls()

    Foo()

    assert Foo.__doc__ == "Foo class docstring for testing."
    assert len(recwarn) == 1
    assert recwarn[0].category == GlotaranApiDeprecationWarning
    assert recwarn[0].message.args[0] == DEPRECATION_WARN_MESSAGE
    assert Path(recwarn[0].filename) == Path(__file__)

    Foo.from_string("foo")

    assert len(recwarn) == 2


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_dict_key_swap_keys():
    """Replace old with new key while keeping the value."""
    test_dict = {"foo": 123}
    with pytest.warns(
        GlotaranApiDeprecationWarning, match="'foo'.+was deprecated, use 'bar'"
    ) as record:
        deprecate_dict_entry(
            dict_to_check=test_dict,
            deprecated_usage="foo",
            new_usage="bar",
            to_be_removed_in_version="0.6.0",
            swap_keys=("foo", "bar"),
        )

        assert "bar" in test_dict
        assert test_dict["bar"] == 123
        assert "foo" not in test_dict

        assert len(record) == 1
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_dict_key_replace_rules_only_values():
    """Replace old value for key with new value."""
    test_dict = {"foo": 123}
    with pytest.warns(
        GlotaranApiDeprecationWarning, match="'foo: 123'.+was deprecated, use 'foo: 321'"
    ) as record:
        deprecate_dict_entry(
            dict_to_check=test_dict,
            deprecated_usage="foo: 123",
            new_usage="foo: 321",
            to_be_removed_in_version="0.6.0",
            replace_rules=({"foo": 123}, {"foo": 321}),
        )

        assert "foo" in test_dict
        assert test_dict["foo"] == 321

        assert len(record) == 1
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_dict_key_replace_rules_keys_and_values():
    """Replace old with new key AND replace old value for key with new value."""
    test_dict = {"foo": 123}
    with pytest.warns(
        GlotaranApiDeprecationWarning, match="'foo: 123'.+was deprecated, use 'bar: 321'"
    ) as record:
        deprecate_dict_entry(
            dict_to_check=test_dict,
            deprecated_usage="foo: 123",
            new_usage="bar: 321",
            to_be_removed_in_version="0.6.0",
            replace_rules=({"foo": 123}, {"bar": 321}),
        )

        assert "bar" in test_dict
        assert test_dict["bar"] == 321
        assert "foo" not in test_dict

        assert len(record) == 1
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.xfail(strict=True)
@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_dict_key_does_not_apply_swap_keys():
    """Don't warn if the dict doesn't change because old_key didn't match"""

    with pytest.warns(
        GlotaranApiDeprecationWarning, match="'foo: 123'.+was deprecated, use 'foo: 321'"
    ):
        deprecate_dict_entry(
            dict_to_check={"foo": 123},
            deprecated_usage="foo: 123",
            new_usage="foo: 321",
            to_be_removed_in_version="0.6.0",
            swap_keys=("bar", "baz"),
        )


@pytest.mark.xfail(strict=True)
@pytest.mark.parametrize(
    "replace_rules",
    (
        ({"bar": 123}, {"bar": 321}),
        ({"foo": 111}, {"bar": 321}),
    ),
)
@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_dict_key_does_not_apply(
    replace_rules: tuple[Mapping[Hashable, Any], Mapping[Hashable, Any]]
):
    """Don't warn if the dict doesn't change because old_key or old_value didn't match"""
    with pytest.warns(
        GlotaranApiDeprecationWarning, match="'foo: 123'.+was deprecated, use 'foo: 321'"
    ):
        deprecate_dict_entry(
            dict_to_check={"foo": 123},
            deprecated_usage="foo: 123",
            new_usage="foo: 321",
            to_be_removed_in_version="0.6.0",
            replace_rules=replace_rules,
        )


@pytest.mark.parametrize(
    "swap_keys, replace_rules",
    (
        (None, None),
        (("bar", "baz"), ({"bar": 1}, {"baz": 2})),
    ),
)
@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_dict_key_error_no_action(
    swap_keys: tuple[Hashable, Hashable] | None,
    replace_rules: tuple[Mapping[Hashable, Any], Mapping[Hashable, Any]] | None,
):
    """Raise error if none or both `swap_keys` and `replace_rules` were provided."""
    with pytest.raises(
        ValueError,
        match=(
            r"Exactly one of the parameters `swap_keys` or `replace_rules` needs to be provided\."
        ),
    ):
        deprecate_dict_entry(
            dict_to_check={},
            deprecated_usage="",
            new_usage="",
            to_be_removed_in_version="",
            swap_keys=swap_keys,
            replace_rules=replace_rules,
        )


def test_module_attribute():
    """Same code as the original import"""

    result = module_attribute("glotaran.deprecation.deprecation_utils", "parse_version")

    assert result.__code__ == parse_version.__code__


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_module_attribute():
    """Same code as the original import and warning"""

    with pytest.warns(GlotaranApiDeprecationWarning) as record:
        from glotaran.deprecation.test.dummy_package.deprecated_module_attribute import (
            deprecated_attribute,
        )

        assert deprecated_attribute.__code__ == parse_version.__code__
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_module_attribute_overwrite():
    """Qualname was only used for the warning"""

    with pytest.warns(GlotaranApiDeprecationWarning) as record:
        from glotaran.deprecation.test.dummy_package.deprecated_module_attribute import foo_bar

        assert foo_bar.__code__ == parse_version.__code__
        assert Path(record[0].filename) == Path(__file__)
        assert "glotaran.does_not._need_to_exists" in str(record[0].message)


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_submodule(recwarn: WarningsRecorder):
    """Raise warning when Attribute of fake module is used"""

    from glotaran.deprecation.test.dummy_package import deprecated_module

    assert deprecated_module.parse_version.__code__ == parse_version.__code__

    assert len(recwarn) == 1
    assert recwarn[0].category == GlotaranApiDeprecationWarning


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_submodule_from_import(recwarn: WarningsRecorder):
    """Raise warning when Attribute of fake module is imported"""

    from glotaran.deprecation.test.dummy_package.deprecated_module import (  # noqa: F401
        parse_version,
    )

    assert len(recwarn) == 1
    assert recwarn[0].category == GlotaranApiDeprecationWarning
    assert Path(recwarn[0].filename) == Path(__file__)


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_submodule_from_import_overwrite(recwarn: WarningsRecorder):
    """Qualname was only used for the warning"""

    from glotaran.deprecation.test.dummy_package.overwritten_module import (  # noqa: F401
        parse_version,
    )

    assert len(recwarn) == 1
    assert recwarn[0].category == GlotaranApiDeprecationWarning
    assert Path(recwarn[0].filename) == Path(__file__)
    assert "glotaran.does_not._need_to_exists" in str(recwarn[0].message)


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_submodule_import_error(recwarn: WarningsRecorder):
    """Raise warning when Attribute of fake module is imported"""

    with pytest.raises(ImportError) as excinfo:
        from glotaran.deprecation.test.dummy_package.deprecated_module import (  # noqa: F401
            does_not_exists,
        )

    assert str(excinfo.value) == (
        "cannot import name 'does_not_exists' from "
        "'glotaran.deprecation.test.dummy_package.deprecated_module' (unknown location)"
    )


@pytest.mark.usefixtures("glotaran_0_3_0")
def test_deprecate_submodule_attr__file__(recwarn: WarningsRecorder):
    """Now warning when inspecting __file__ attribute (pytest using inspect)"""
    warnings.simplefilter("always")

    from glotaran.deprecation.test import dummy_package

    dummy_package.__file__

    assert len(recwarn) == 0



from glotaran.builtin.megacomplexes.decay.decay_megacomplex import DecayMegacomplex
from glotaran.builtin.megacomplexes.decay.decay_parallel_megacomplex import (
    DecayParallelMegacomplex,
)
from glotaran.builtin.megacomplexes.decay.decay_sequential_megacomplex import (
    DecaySequentialMegacomplex,
)



"""Helper functions to give deprecation warnings."""

from __future__ import annotations

import os
import re
import sys
from collections.abc import Callable
from collections.abc import Hashable
from collections.abc import Mapping
from collections.abc import MutableMapping
from functools import wraps
from importlib import import_module
from importlib.metadata import distribution
from types import ModuleType
from typing import TYPE_CHECKING
from typing import Any
from typing import TypeVar
from typing import cast
from warnings import warn

import numpy as np

DecoratedCallable = TypeVar(
    "DecoratedCallable", bound=Callable[..., Any]
)  # decorated function or class

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import NoReturn


class OverDueDeprecation(Exception):
    """Error thrown when a deprecation should have been removed.

    See Also
    --------
    deprecate
    warn_deprecated
    deprecate_module_attribute
    deprecate_submodule
    deprecate_dict_entry
    """


class GlotaranDeprecatedApiError(Exception):
    """Exception raised when a deprecation has no replacement.

    See Also
    --------
    deprecate
    warn_deprecated
    deprecate_module_attribute
    deprecate_submodule
    deprecate_dict_entry
    """


class GlotaranApiDeprecationWarning(UserWarning):
    """Warning to give users about API changes.

    See Also
    --------
    deprecate
    warn_deprecated
    deprecate_module_attribute
    deprecate_submodule
    deprecate_dict_entry
    """


def glotaran_version() -> str:
    """Version of the distribution.

    This is basically the same as ``glotaran.__version__`` but independent from glotaran.
    This way all of the deprecation functionality can be used even in
    ``glotaran.__init__.py`` without moving the import below the definition of
    ``__version__`` or causeing a circular import issue.

    Returns
    -------
    str
        The version string.
    """
    return distribution("pyglotaran").version


def parse_version(version_str: str) -> tuple[int, int, int]:
    """Parse version string to tuple of three ints for comparison.

    Parameters
    ----------
    version_str : str
        Fully qualified version string of the form 'major.minor.patch'.

    Returns
    -------
    tuple[int, int, int]
        Version as tuple.

    Raises
    ------
    ValueError
        If ``version_str`` has less that three elements separated by ``.``.
    ValueError
        If ``version_str`` 's first three elements can not be casted to int.
    """
    error_message = (
        "version_str needs to be a fully qualified version consisting of "
        f"int parts (e.g. '0.0.1'), got {version_str!r}"
    )
    split_version = version_str.partition("-")[0].split(".")
    if len(split_version) < 3:
        raise ValueError(error_message)
    try:
        return tuple(
            map(int, (*split_version[:2], split_version[2].partition("rc")[0]))
        )  # type:ignore[return-value]
    except ValueError as err:
        raise ValueError(error_message) from err


def check_qualnames_in_tests(qual_names: Sequence[str], importable_indices: Sequence[int]):
    """Test that qualnames import path exists when running tests.

    All deprecations should be tested anyway in order to get the proper
    errors when a deprecation is overdue.
    This helperfunction also helps to ensure that at least the import
    paths (``qual_names``) of the old and new usage exist.

    Parameters
    ----------
    qual_names : Sequence[str]
        Sequence of fully qualified module attribute names,
        optionally with call arguments.
    importable_indices: Sequence[int]
        Indices of corresponding to ``qual_names`` indicating
        how to slice each ``qual_name`` split at ``.``, for the import
        and attribute checking.

    See Also
    --------
    warn_deprecated
    deprecate
    """
    # Since this is always true for tests run with pytest we ignore the branch coverage
    if "PYTEST_CURRENT_TEST" in os.environ:  # pragma: no branch
        for qual_name, slice_index in zip(qual_names, importable_indices):
            qual_name_parts = qual_name.partition("(")[0].partition("[")[0].split(".")
            module_name = ".".join(qual_name_parts[:-slice_index])
            object_name = qual_name_parts[-slice_index]
            module = __import__(module_name, fromlist=(object_name))
            assert hasattr(module, object_name)
            if slice_index != 1:
                item = getattr(module, object_name)
                hasattr(item, qual_name_parts[-slice_index + 1])


def check_overdue(deprecated_qual_name_usage: str, to_be_removed_in_version: str) -> None:
    """Check if a deprecation is overdue for removal.

    Parameters
    ----------
    deprecated_qual_name_usage : str
        Old usage with fully qualified name e.g.:
        ``'glotaran.read_model_from_yaml(model_yml_str)'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.

    Raises
    ------
    OverDueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.
    """
    if (
        parse_version(glotaran_version()) >= parse_version(to_be_removed_in_version)
        and "dev" not in glotaran_version()
    ):
        raise OverDueDeprecation(
            f"Support for {deprecated_qual_name_usage.partition('(')[0]!r} was "
            f"supposed to be dropped in version: {to_be_removed_in_version!r}.\n"
            f"Current version is: {glotaran_version()!r}"
        )


def raise_deprecation_error(
    *,
    deprecated_qual_name_usage: str,
    new_qual_name_usage: str,
    to_be_removed_in_version: str,
) -> NoReturn:
    """Raise :class:`GlotaranDeprectedApiError` error, with formatted message.

    This should only be used if there is no reasonable way to keep the deprecated
    usage functional!

    Parameters
    ----------
    deprecated_qual_name_usage : str
        Old usage with fully qualified name e.g.:
        ``'glotaran.read_model_from_yaml(model_yml_str)'``
    new_qual_name_usage : str
        New usage as fully qualified name e.g.:
        ``'glotaran.io.load_model(model_yml_str, format_name="yml_str")'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.

    Raises
    ------
    OverDueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.
    GlotaranDeprecatedApiError
        If :class:`OverDueDeprecation` wasn't raised before.


    .. # noqa: DAR402 OverDueDeprecation
    .. # noqa: DAR401 GlotaranDeprecatedApiError
    """
    check_overdue(deprecated_qual_name_usage, to_be_removed_in_version)
    message = (
        f"Usage of {deprecated_qual_name_usage!r} was deprecated, "
        f"use {new_qual_name_usage!r} instead.\n"
        "It wasn't possible to restore the original behavior of this usage "
        "(most likely due to an object hierarchy change)."
        "This usage change message won't be show as of version: "
        f"{to_be_removed_in_version!r}."
    )
    raise GlotaranDeprecatedApiError(message)


def warn_deprecated(
    *,
    deprecated_qual_name_usage: str,
    new_qual_name_usage: str,
    to_be_removed_in_version: str,
    check_qual_names: tuple[bool, bool] = (True, True),
    stacklevel: int = 2,
    importable_indices: tuple[int, int] = (1, 1),
) -> None:
    """Raise deprecation warning with change information.

    The change information are old / new usage information and end of support version.

    Parameters
    ----------
    deprecated_qual_name_usage : str
        Old usage with fully qualified name e.g.:
        ``'glotaran.read_model_from_yaml(model_yml_str)'``
    new_qual_name_usage : str
        New usage as fully qualified name e.g.:
        ``'glotaran.io.load_model(model_yml_str, format_name="yml_str")'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.
    check_qual_names : tuple[bool, bool]
        Whether or not to check for the existence ``deprecated_qual_name_usage`` and
        ``deprecated_qual_name_usage``

        *   Set the first value to False to prevent infinite recursion error when changing
            a module attribute import.

        *   Set the second value to False if the new usage in in a different package or
            there is none.

    stacklevel: int
        Stack at which the warning should be shown as raise. Default: 2

    importable_indices : tuple[int, int]
        Indices from right for most nested item which is importable for
        ``deprecated_qual_name_usage`` and ``new_qual_name_usage``
        after splitting at ``.``. This is used when the old or new usage
        is a method or mapping access. E.g. let ``deprecated_qual_name_usage``
        be ``package.module.class.mapping["key"]``, then you would use
        ``importable_indices=(2, 1)``, this way func:`check_qualnames_in_tests`
        will import ``package.module.class`` and check if ``class`` has an attribute
        ``mapping``.

    Raises
    ------
    OverDueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.

    See Also
    --------
    deprecate
    deprecate_module_attribute
    deprecate_submodule
    check_qualnames_in_tests

    Examples
    --------
    This is the way the old ``read_parameters_from_yaml_file`` could deprecated and the usage of
    ``load_model`` being promoted instead.


    .. code-block:: python
        :caption: glotaran/deprecation/modules/glotaran_root.py

        def read_parameters_from_yaml_file(model_path: str):
            warn_deprecated(
                deprecated_qual_name_usage="glotaran.read_parameters_from_yaml_file(model_path)",
                new_qual_name_usage="glotaran.io.load_model.load_model(model_path)",
                to_be_removed_in_version="0.6.0",
            )
            return load_model(model_path)


    .. # noqa: DAR402
    """
    check_overdue(deprecated_qual_name_usage, to_be_removed_in_version)
    qual_names = (deprecated_qual_name_usage, new_qual_name_usage)
    selected_qual_names = [
        qual_name for qual_name, check in zip(qual_names, check_qual_names) if check
    ]
    selected_indices = importable_indices[: len(selected_qual_names)]
    check_qualnames_in_tests(qual_names=selected_qual_names, importable_indices=selected_indices)
    warn(
        GlotaranApiDeprecationWarning(
            f"Usage of {deprecated_qual_name_usage!r} was deprecated, "
            f"use {new_qual_name_usage!r} instead.\n"
            f"This usage will be an error in version: {to_be_removed_in_version!r}."
        ),
        stacklevel=stacklevel,
    )


def deprecate(
    *,
    deprecated_qual_name_usage: str,
    new_qual_name_usage: str,
    to_be_removed_in_version: str,
    has_glotaran_replacement: bool = True,
    importable_indices: tuple[int, int] = (1, 1),
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """Decorate a function, method or class to deprecate it.

    This raises deprecation warning with old / new usage information and
    end of support version.


    Parameters
    ----------
    deprecated_qual_name_usage : str
        Old usage with fully qualified name e.g.:
        ``'glotaran.read_model_from_yaml(model_yml_str)'``
    new_qual_name_usage : str
        New usage as fully qualified name e.g.:
        ``'glotaran.io.load_model(model_yml_str, format_name="yml_str")'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.
    has_glotaran_replacement : bool
        Whether or not this functionality has a replacement in core
        pyglotaran. This will be mapped to the second entry of ``check_qualnames``
        in :func:`warn_deprecated`.
    importable_indices : Sequence[int]
        Indices from right for most nested item which is importable for
        ``deprecated_qual_name_usage`` and ``new_qual_name_usage``
        after splitting at ``.``. This is used when the old or new usage
        is a method or mapping access. E.g. let ``deprecated_qual_name_usage``
        be ``package.module.class.mapping["key"]``, then you would use
        ``importable_indices=(2, 1)``, this way func:`check_qualnames_in_tests`
        will import ``package.module.class`` and check if ``class`` has an attribute
        ``mapping``. Default


    Returns
    -------
    DecoratedCallable
        Original function or class throwing a Deprecation warning when used.

    Raises
    ------
    OverDueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.

    See Also
    --------
    warn_deprecated
    deprecate_module_attribute
    deprecate_submodule
    check_qualnames_in_tests

    Examples
    --------
    This is the way the old ``read_parameters_from_yaml_file`` was deprecated and the usage of
    ``load_model`` was promoted instead.


    .. code-block:: python
        :caption: glotaran/deprecation/modules/glotaran_root.py

        @deprecate(
            deprecated_qualname_usage="glotaran.read_parameters_from_yaml_file(model_path)",
            new_qualname_usage="glotaran.io.load_model(model_path)",
            to_be_removed_in_version="0.6.0",
        )
        def read_parameters_from_yaml_file(model_path: str):
            return load_model(model_path)


    .. # noqa: DAR402
    """

    def inject_warn_into_call(deprecated_object: DecoratedCallable) -> DecoratedCallable:
        """Wrap warning into function call.

        Used on deprecated_object.__new__ if it's a class or else on deprecated_object.
        """

        @wraps(deprecated_object)
        def inner_wrapper(*args: Any, **kwargs: Any) -> DecoratedCallable:
            """Wrap running the function and warning."""
            warn_deprecated(
                deprecated_qual_name_usage=deprecated_qual_name_usage,
                new_qual_name_usage=new_qual_name_usage,
                to_be_removed_in_version=to_be_removed_in_version,
                stacklevel=3,
                check_qual_names=(True, has_glotaran_replacement),
                importable_indices=importable_indices,
            )
            return deprecated_object(*args, **kwargs)

        return cast(DecoratedCallable, inner_wrapper)

    def outer_wrapper(deprecated_object: DecoratedCallable) -> DecoratedCallable:
        """Wrap deprecated_object of all callable kinds."""
        if not isinstance(deprecated_object, type):
            return cast(DecoratedCallable, inject_warn_into_call(deprecated_object))

        setattr(
            deprecated_object,
            "__new__",
            inject_warn_into_call(deprecated_object.__new__),
        )
        return deprecated_object  # type: ignore[return-value]

    return cast(Callable[[DecoratedCallable], DecoratedCallable], outer_wrapper)


def deprecate_dict_entry(
    *,
    dict_to_check: MutableMapping[Hashable, Any],
    deprecated_usage: str,
    new_usage: str,
    to_be_removed_in_version: str,
    swap_keys: tuple[Hashable, Hashable] | None = None,
    replace_rules: tuple[Mapping[Hashable, Any], Mapping[Hashable, Any]] | None = None,
    stacklevel: int = 3,
) -> None:
    """Replace dict entry inplace and warn about usage change, if present in the dict.

    Parameters
    ----------
    dict_to_check : MutableMapping[Hashable, Any]
        Dict which should be checked.
    deprecated_usage : str
        Old usage to inform user (only used in warning).
    new_usage : str
        New usage to inform user (only used in warning).
    to_be_removed_in_version : str
        Version the support for this usage will be removed.
    swap_keys : tuple[Hashable, Hashable]
        (old_key, new_key),
        ``dict_to_check[new_key]`` will be assigned the value ``dict_to_check[old_key]``
        and ``old_key`` will be removed from the dict.
        by default None
    replace_rules : Mapping[Hashable, tuple[Any, Any]]
        ({old_key: old_value}, {new_key: new_value}),
        If ``dict_to_check[old_key]`` has the value ``old_value``,
        ``dict_to_check[new_key]`` it will be set to ``new_value``.
        ``old_key`` will be removed from the dict if ``old_key`` and ``new_key`` aren't equal.
        by default None
    stacklevel : int
        Stack at which the warning should be shown as raise. , by default 3


    Raises
    ------
    ValueError
        If both ``swap_keys`` and ``replace_rules`` are None (default) or not None.
    OverDueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.

    See Also
    --------
    warn_deprecated

    Notes
    -----
    To prevent confusion exactly one of ``replace_rules`` and ``swap_keys``
    needs to be passed.

    Examples
    --------
    For readability sake the warnings won't be shown in the examples.

    Swapping key names:

    >>> dict_to_check = {"foo": 123}
    >>> deprecate_dict_entry(
            dict_to_check=dict_to_check,
            deprecated_usage="foo",
            new_usage="bar",
            to_be_removed_in_version="0.6.0",
            swap_keys=("foo", "bar")
        )
    >>> dict_to_check
    {"bar": 123}

    Changing values:

    >>> dict_to_check = {"foo": 123}
    >>> deprecate_dict_entry(
            dict_to_check=dict_to_check,
            deprecated_usage="foo: 123",
            new_usage="foo: 123.0",
            to_be_removed_in_version="0.6.0",
            replace_rules=({"foo": 123}, {"foo": 123.0})
        )
    >>> dict_to_check
    {"foo": 123.0}

    Swapping key names AND changing values:

    >>> dict_to_check = {"type": "kinetic-spectrum"}
    >>> deprecate_dict_entry(
            dict_to_check=dict_to_check,
            deprecated_usage="type: kinectic-spectrum",
            new_usage="default_megacomplex: decay",
            to_be_removed_in_version="0.6.0",
            replace_rules=({"type": "kinetic-spectrum"}, {"default_megacomplex": "decay"})
        )
    >>> dict_to_check
    {"default_megacomplex": "decay"}


    .. # noqa: DAR402
    """
    dict_changed = False

    if not np.logical_xor(swap_keys is None, replace_rules is None):
        raise ValueError(
            "Exactly one of the parameters `swap_keys` or `replace_rules` needs to be provided."
        )
    if swap_keys is not None and swap_keys[0] in dict_to_check:
        dict_changed = True
        dict_to_check[swap_keys[1]] = dict_to_check[swap_keys[0]]
        del dict_to_check[swap_keys[0]]
    if replace_rules is not None:
        old_key, old_value = next(iter(replace_rules[0].items()))
        new_key, new_value = next(iter(replace_rules[1].items()))
        if old_key in dict_to_check and dict_to_check[old_key] == old_value:
            dict_changed = True
            dict_to_check[new_key] = new_value
            if new_key != old_key:
                del dict_to_check[old_key]

    if dict_changed:
        warn_deprecated(
            deprecated_qual_name_usage=deprecated_usage,
            new_qual_name_usage=new_usage,
            to_be_removed_in_version=to_be_removed_in_version,
            stacklevel=stacklevel,
            check_qual_names=(False, False),
        )


def module_attribute(module_qual_name: str, attribute_name: str) -> Any:
    """Import and return the attribute (e.g. function or class) of a module.

    This is basically the same as ``from module_name import attribute_name as return_value``
    where this function returns ``return_value``.

    Parameters
    ----------
    module_qual_name : str
        Fully qualified name for a module e.g. ``glotaran.model.base_model``
    attribute_name : str
        Name of the attribute e.g. ``Model``

    Returns
    -------
    Any
        Attribute of the module, e.g. a function or class.
    """
    module = import_module(module_qual_name)
    return getattr(module, attribute_name)


def deprecate_module_attribute(
    *,
    deprecated_qual_name: str,
    new_qual_name: str,
    to_be_removed_in_version: str,
    module_load_overwrite: str = "",
) -> Any:
    """Import and return and anttribute from the new location.

    This needs to be wrapped in the definition of a module wide
    ``__getattr__`` function so it won't throw warnings all the time
    (see example).

    Parameters
    ----------
    deprecated_qual_name : str
        Fully qualified name of the deprecated attribute e.g.:
        ``glotaran.ParameterGroup``
    new_qual_name : str
        Fully qualified name of the new attribute e.g.:
        ``glotaran.parameter.ParameterGroup``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.
    module_load_overwrite : str
        Overwrite the location the functionality will be set from.
        This allows preserving functionality without polluting a new
        module with code just for the sake of it. By default ''

    Returns
    -------
    Any
        Module attribute from its new location.

    Raises
    ------
    OverDueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.

    See Also
    --------
    deprecate
    warn_deprecated
    deprecate_submodule

    Examples
    --------
    When deprecating the usage of ``ParameterGroup`` the root of ``glotaran``
    and promoting to import it from ``glotaran.parameter`` the following code
    was added to the root ``__init__.py``.

    .. code-block:: python
        :caption: glotaran/__init__.py


        def __getattr__(attribute_name: str):
            from glotaran.deprecation import deprecate_module_attribute

            if attribute_name == "ParameterGroup":
                return deprecate_module_attribute(
                    deprecated_qual_name="glotaran.ParameterGroup",
                    new_qual_name="glotaran.parameter.ParameterGroup",
                    to_be_removed_in_version="0.6.0",
                )

            raise AttributeError(f"module {__name__} has no attribute {attribute_name}")


    .. # noqa: DAR402
    """
    if not module_load_overwrite:
        module_name = ".".join(new_qual_name.split(".")[:-1])
        attribute_name = new_qual_name.split(".")[-1]
        check_qual_names = (False, True)
    else:
        module_name = ".".join(module_load_overwrite.split(".")[:-1])
        attribute_name = module_load_overwrite.split(".")[-1]
        check_qual_names = (False, False)

    warn_deprecated(
        deprecated_qual_name_usage=re.sub(r"\.__path__$", "", deprecated_qual_name),
        new_qual_name_usage=re.sub(r"\.__path__$", "", new_qual_name),
        to_be_removed_in_version=to_be_removed_in_version,
        check_qual_names=check_qual_names,
        stacklevel=4,
        importable_indices=(1, 1),
    )
    return module_attribute(module_name, attribute_name)


def deprecate_submodule(
    *,
    deprecated_module_name: str,
    new_module_name: str,
    to_be_removed_in_version: str,
    module_load_overwrite: str = "",
) -> ModuleType:
    r"""Create a module at runtime which retrieves attributes from new module.

    When moving a module, create a variable with the modules name in the
    parent packages ``__init__.py``, so imports will be redirected to the
    new module location and a deprecation warning will be given, to help
    the user adjust the outdated code.
    Each time an attribute is retrieved there will be a deprecation warning.

    Parameters
    ----------
    deprecated_module_name : str
        Fully qualified name of the deprecated module e.g.:
        ``'glotaran.analysis.result'``
    new_module_name : str
        Fully qualified name of the new module e.g.:
        ``'glotaran.project.result'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.
    module_load_overwrite : str
        Overwrite the location for the new module the deprecated functionality is loaded from.
        This allows preserving functionality without polluting a new
        module with code just for the sake of it. By default ''

    Returns
    -------
    ModuleType
        Module containing

    Raises
    ------
    OverDueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.

    See Also
    --------
    deprecate
    deprecate_module_attribute

    Examples
    --------
    When moving the module ``result`` from ``glotaran.analysis.result`` to
    ``glotaran.project.result`` the following code was added to the old parent
    packages (``glotaran.analysis``) ``__init__.py``.

    .. code-block:: python
        :caption: glotaran/analysis/__init__.py

        from glotaran.deprecation.deprecation_utils import deprecate_submodule

        result = deprecate_submodule(
            deprecated_module_name="glotaran.analysis.result",
            new_module_name="glotaran.project.result",
            to_be_removed_in_version="0.6.0",
        )


    .. # noqa: DAR402
    """
    new_module = (
        import_module(module_load_overwrite)
        if module_load_overwrite
        else import_module(new_module_name)
    )

    deprecated_module = ModuleType(
        deprecated_module_name,
        f"Deprecated use {new_module_name!r} instead.\n\n{new_module.__doc__}",
    )

    def warn_getattr(attribute_name: str):
        if attribute_name == "__file__":
            return new_module.__file__

        elif attribute_name in dir(new_module):
            return deprecate_module_attribute(
                deprecated_qual_name=f"{deprecated_module_name}.{attribute_name}",
                new_qual_name=f"{new_module_name}.{attribute_name}",
                to_be_removed_in_version=to_be_removed_in_version,
                module_load_overwrite=module_load_overwrite
                and f"{module_load_overwrite}.{attribute_name}",
            )

        raise AttributeError(f"module {deprecated_module_name} has no attribute {attribute_name}")

    setattr(deprecated_module, "__getattr__", warn_getattr)
    setattr(deprecated_module, "__package__", deprecated_module_name.split(".")[:-1])
    setattr(deprecated_module, "__dir__", new_module.__dir__)

    sys.modules[deprecated_module_name] = deprecated_module
    return deprecated_module



"""Module containing the optimization group class."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.io.prepare_dataset import add_svd_to_dataset
from glotaran.model import DatasetGroup
from glotaran.model.dataset_model import finalize_dataset_model
from glotaran.optimization.data_provider import DataProvider
from glotaran.optimization.data_provider import DataProviderLinked
from glotaran.optimization.estimation_provider import EstimationProvider
from glotaran.optimization.estimation_provider import EstimationProviderLinked
from glotaran.optimization.estimation_provider import EstimationProviderUnlinked
from glotaran.optimization.matrix_provider import MatrixProvider
from glotaran.optimization.matrix_provider import MatrixProviderLinked
from glotaran.optimization.matrix_provider import MatrixProviderUnlinked
from glotaran.parameter import Parameters
from glotaran.project import Scheme

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


class OptimizationGroup:
    """A class to optimize a dataset group."""

    def __init__(
        self,
        scheme: Scheme,
        dataset_group: DatasetGroup,
    ):
        """Initialize an optimization group for a dataset group.

        Parameters
        ----------
        scheme : Scheme
            The optimization scheme.
        dataset_group : DatasetGroup
            The dataset group.
        """
        self._dataset_group = dataset_group
        self._dataset_group.set_parameters(scheme.parameters)
        self._data = scheme.data
        self._add_svd = scheme.add_svd
        link_clp = dataset_group.link_clp
        if link_clp is None:
            link_clp = dataset_group.is_linkable(scheme.parameters, scheme.data)

        if link_clp:
            data_provider = DataProviderLinked(scheme, dataset_group)
            matrix_provider = MatrixProviderLinked(dataset_group, data_provider)
            estimation_provider = EstimationProviderLinked(
                dataset_group, data_provider, matrix_provider
            )
        else:
            data_provider = DataProvider(scheme, dataset_group)  # type:ignore[assignment]
            matrix_provider = MatrixProviderUnlinked(  # type:ignore[assignment]
                self._dataset_group, data_provider
            )
            estimation_provider = EstimationProviderUnlinked(  # type:ignore[assignment]
                dataset_group, data_provider, matrix_provider  # type:ignore[arg-type]
            )

        self._data_provider: DataProvider = data_provider
        self._matrix_provider: MatrixProvider = matrix_provider
        self._estimation_provider: EstimationProvider = estimation_provider

        if self._add_svd:
            for dataset in self._data.values():
                self.add_svd_data(
                    "data",
                    dataset,
                    dataset.data.dims[0],
                    dataset.data.dims[1],
                )

    def calculate(self, parameters: Parameters):
        """Calculate the optimization group data.

        Parameters
        ----------
        parameters : Parameters
            The parameters.
        """
        self._dataset_group.set_parameters(parameters)
        self._matrix_provider.calculate()
        self._estimation_provider.estimate()

    def get_additional_penalties(self) -> list[float]:
        """Get additional penalties.

        Returns
        -------
        list[float]
            The additional penalties.
        """
        return self._estimation_provider.get_additional_penalties()

    def get_full_penalty(self) -> ArrayLike:
        """Get the full penalty.

        Returns
        -------
        ArrayLike
            The full penalty.
        """
        return self._estimation_provider.get_full_penalty()

    def add_weight_to_result_data(self, dataset_label: str, result_dataset: xr.Dataset):
        """Add weight to result dataset if dataset is weighted.

        Parameters
        ----------
        dataset_label : str
            The label of the data.
        result_dataset : xr.Dataset
            The label of the data.
        """
        weight = self._data_provider.get_weight(dataset_label)
        if weight is None:
            return
        result_dataset["weighted_residual"] = result_dataset["residual"]
        result_dataset["residual"] = result_dataset["residual"] / weight
        if "weight" not in result_dataset:
            if weight.shape != result_dataset.data.shape:
                weight = weight.T
            result_dataset["weight"] = (result_dataset.data.dims, weight)

    def create_result_data(self) -> dict[str, xr.Dataset]:
        """Create resulting datasets.

        Returns
        -------
        dict[str, xr.Dataset]
            The datasets with the results.
        """
        result_datasets = {
            label: data.copy()
            for label, data in self._data.items()
            if label in self._dataset_group.dataset_models.keys()
        }

        global_matrices, matrices = self._matrix_provider.get_result()
        clps, residuals = self._estimation_provider.get_result()

        for label, dataset_model in self._dataset_group.dataset_models.items():
            result_dataset = result_datasets[label]

            model_dimension = self._data_provider.get_model_dimension(label)
            result_dataset.attrs["model_dimension"] = model_dimension
            global_dimension = self._data_provider.get_global_dimension(label)
            result_dataset.attrs["global_dimension"] = global_dimension

            result_dataset["residual"] = residuals[label]
            self.add_weight_to_result_data(label, result_dataset)

            result_dataset["matrix"] = matrices[label]
            if label in global_matrices:
                result_dataset["global_matrix"] = global_matrices[label]
            result_dataset["clp"] = clps[label]

            if self._add_svd:
                self.add_svd_data("residual", result_dataset, model_dimension, global_dimension)
                if "weighted_residual" in result_dataset:
                    self.add_svd_data(
                        "weighted_residual", result_dataset, model_dimension, global_dimension
                    )

            # Calculate RMS
            size = result_dataset.residual.shape[0] * result_dataset.residual.shape[1]
            result_dataset.attrs["root_mean_square_error"] = np.sqrt(
                (result_dataset.residual**2).sum() / size
            ).data

            result_dataset.attrs["weighted_root_mean_square_error"] = (
                np.sqrt((result_dataset.weighted_residual**2).sum() / size).data
                if "weighted_residual" in result_dataset
                else result_dataset.attrs["root_mean_square_error"]
            )

            result_dataset.attrs["dataset_scale"] = (
                1
                if dataset_model.scale is None
                else dataset_model.scale.value  # type:ignore[union-attr]
            )

            # reconstruct fitted data
            result_dataset["fitted_data"] = result_dataset.data - result_dataset.residual

            finalize_dataset_model(dataset_model, result_dataset)

        return result_datasets

    @staticmethod
    def add_svd_data(name: str, dataset: xr.Dataset, lsv_dim: str, rsv_dim: str):
        """Add the SVD of a data matrix to a dataset.

        Parameters
        ----------
        name : str
            Name of the data matrix.
        dataset : xr.Dataset
            Dataset containing the data, which will be updated with the SVD values.
        lsv_dim : str
            The dimension name of the left singular vectors.
        rsv_dim : str
            The dimension name of the right singular vectors.
        """
        add_svd_to_dataset(
            dataset, name=name, lsv_dim=lsv_dim, rsv_dim=rsv_dim, data_array=dataset[name]
        )

    @property
    def number_of_clps(self) -> int:
        """Return number of conditionally linear parameters.

        Returns
        -------
        int
        """
        return self._matrix_provider.number_of_clps



"""Glotaran module with utilities for sanitation of parsed content."""
from __future__ import annotations

from typing import Any

import numpy as np

from glotaran.utils.regex import RegexPattern as rp


def pretty_format_numerical(value: float | int, decimal_places: int = 1) -> str:
    """Format value with with at most ``decimal_places`` decimal places.

    Used to format values like the t-value.

    Parameters
    ----------
    value: float | int
        Numerical value to format.
    decimal_places: int
        Decimal places to display. Defaults to 1

    Returns
    -------
    str
        Pretty formatted version of the value.
    """
    # Bool returned by numpy do not support the ``is`` comparison (not same singleton as in python)
    # Ref: https://stackoverflow.com/a/37744300/3990615
    if not np.isfinite(value):
        return str(value)
    if abs(value - int(value)) <= np.finfo(np.float64).eps:
        return str(int(value))
    abs_value = abs(value)
    if abs_value < 10 ** (-decimal_places):
        format_instruction = f".{decimal_places}e"
    elif abs_value < 10 ** (decimal_places):
        format_instruction = f".{decimal_places}f"
    else:
        format_instruction = ".0f"
    return f"{value:{format_instruction}}"


def sanitize_list_with_broken_tuples(mangled_list: list[str | float]) -> list[str]:
    """Sanitize a list with 'broken' tuples.

    A list of broken tuples as returned by yaml when parsing tuples.
    e.g parsing the list of tuples [(3,100), (4,200)] results in
    a list of str ['(3', '100)', '(4', '200)'] which can be restored to
    a list with the tuples restored as strings ['(3, 100)', '(4, 200)']

    Parameters
    ----------
    mangled_list : List[Union[str,float]]
        A list with strings representing tuples broken up by round brackets.

    Returns
    -------
    List[str]
        A list containing the restores tuples (in string form) which can be
        converted back to numbered tuples using `list_string_to_tuple`
    """
    sanitized_string = str(mangled_list).replace("'", "")
    return list(rp.elements_in_string_of_list.findall(sanitized_string))


def sanitize_dict_keys(d: dict) -> dict:
    """Sanitize the stringified tuple dict keys in a yaml parsed dict.

    Keys representing a tuple, e.g. '(s1, s2)' are converted to a tuple of strings
        e.g. ('s1', 's2')

    Parameters
    ----------
    d : dict
        A dict containing tuple-like string keys

    Returns
    -------
    dict
        A dict with tuple-like string keys converted to tuple keys
    """
    if not isinstance(d, (dict, list)):
        return {}
    d_new = {}
    for k, v in d.items() if isinstance(d, dict) else enumerate(d):
        if isinstance(d, dict) and isinstance(k, str) and rp.tuple_word.match(k):
            k_new = tuple(map(str, rp.word.findall(k)))
            d_new[k_new] = v
        elif isinstance(d, (dict, list)):
            if new_v := sanitize_dict_keys(v):
                d[k] = new_v
    return d_new


def sanity_scientific_notation_conversion(d: dict[str, Any] | list[Any]):
    """Convert scientific notation string values to floats.

    Parameters
    ----------
    d : dict[str, Any] | list[Any]
        Iterable which should be checked for scientific notation values.
    """
    if not isinstance(d, (dict, list)):
        return
    for k, v in d.items() if isinstance(d, dict) else enumerate(d):
        if isinstance(v, (list, dict)):
            sanity_scientific_notation_conversion(v)
        if isinstance(v, str):
            d[k] = convert_scientific_to_float(v)  # type: ignore[index,call-overload]


def sanitize_dict_values(d: dict[str, Any] | list[Any]):
    """Sanitizes a dict with broken tuples inside modifying it in-place.

    Broken tuples are tuples that are turned into strings by the yaml parser.
    This functions calls `sanitize_list_with_broken_tuples` to glue the broken strings together
    and then calls list_to_tuple to turn the list with tuple strings back to number tuples.

    Parameters
    ----------
    d : dict
        A (complex) dict containing (possibly nested) values of broken tuple strings.
    """
    if not isinstance(d, (dict, list)):
        return
    for k, v in d.items() if isinstance(d, dict) else enumerate(d):
        if isinstance(v, list):
            leaf = all(isinstance(el, (str, tuple, float)) for el in v)
            if leaf:
                if "(" in str(v):
                    d[k] = list_string_to_tuple(  # type: ignore[index,call-overload]
                        sanitize_list_with_broken_tuples(v)
                    )
            else:
                sanitize_dict_values(v)
        if isinstance(v, dict):
            sanitize_dict_values(v)
        if isinstance(v, str):
            d[k] = string_to_tuple(v)  # type: ignore[index,call-overload]


def string_to_tuple(
    tuple_str: str, from_list=False
) -> tuple[float, ...] | tuple[str, ...] | float | str:
    """Convert a string to a tuple if it matches a tuple pattern.

    Parameters
    ----------
    tuple_str : str
        A string representing some tuple to convert
        the numbers inside the string tuple are mapped to float
    from_list : bool, optional
        only if true will a single number string be converted to float,
        otherwise returned as-is since it may represent a label,
        by default False

    Returns
    -------
    tuple[float], tuple[str], float, str
        Returns the tuple intended by the string
    """
    if rp.tuple_number.match(tuple_str):
        return tuple(map(float, rp.number.findall(tuple_str)))
    elif rp.tuple_word.match(tuple_str):
        return tuple(map(str, rp.word.findall(tuple_str)))
    elif from_list and rp.number.match(tuple_str):
        return float(tuple_str)
    else:
        return tuple_str


def list_string_to_tuple(
    a_list: list[str],
) -> list[tuple[float, ...] | tuple[str, ...] | float | str]:
    """Convert a list of strings (representing tuples) to a list of tuples.

    Parameters
    ----------
    a_list : List[str]
        A list of strings, some of them representing (numbered) tuples

    Returns
    -------
    List[Union[float, str]]
        A list of the (numbered) tuples represted by the incoming a_list
    """
    return [string_to_tuple(v, from_list=True) for v in a_list]


def sanitize_yaml(d: dict, do_keys: bool = True, do_values: bool = False) -> dict:
    """Sanitize a yaml-returned dict for key or (list) values containing tuples.

    Parameters
    ----------
    d : dict
        a dict resulting from parsing a pyglotaran model spec yml file
    do_keys : bool
        toggle sanitization of dict keys, by default True
    do_values : bool
        toggle sanitization of dict values, by default False

    Returns
    -------
    dict
        a sanitized dict with (broken) string tuples restored as proper tuples
    """
    if do_keys:
        sanitize_dict_keys(d)
    if do_values:
        # this is only needed to allow for tuple parsing in specification
        sanitize_dict_values(d)
    sanity_scientific_notation_conversion(d)
    return d


def convert_scientific_to_float(value: str) -> float | str:
    """Convert value to float if it matches scientific notation string.

    Parameters
    ----------
    value : str
        value to convert from string to float if it matches scientific notation

    Returns
    -------
    float | string
        return float if value was scientific notation string, else turn original value
    """
    return float(value) if rp.number_scientific.match(value) else value


def sanitize_parameter_list(parameter_list: list[str | float]) -> list[str | float]:
    """Replace in a list strings matching scientific notation with floats.

    Parameters
    ----------
    parameter_list : list
        A list of parameters where some elements may be strings like 1E7

    Returns
    -------
    list
        A list where strings matching a scientific number have been converted to float
    """
    for i, value in enumerate(parameter_list):
        if isinstance(value, str):
            parameter_list[i] = convert_scientific_to_float(value)

    return parameter_list



from os.path import dirname
from os.path import join

import pytest

from glotaran.builtin.megacomplexes.decay.decay_megacomplex import DecayMegacomplex
from glotaran.builtin.megacomplexes.decay.initial_concentration import InitialConcentration
from glotaran.builtin.megacomplexes.decay.irf import IrfMultiGaussian
from glotaran.builtin.megacomplexes.spectral.shape import SpectralShapeGaussian
from glotaran.io import load_model
from glotaran.model import DatasetModel
from glotaran.model import EqualAreaPenalty
from glotaran.model import OnlyConstraint
from glotaran.model import Weight
from glotaran.model import ZeroConstraint

THIS_DIR = dirname(__file__)


@pytest.fixture
def model():
    spec_path = join(THIS_DIR, "test_model_spec.yml")
    m = load_model(spec_path)
    print(m.markdown())
    return m


def test_dataset(model):
    assert len(model.dataset) == 2

    assert "dataset1" in model.dataset
    dataset = model.dataset["dataset1"]
    assert isinstance(dataset, DatasetModel)
    assert dataset.label == "dataset1"
    assert dataset.megacomplex == ["cmplx1"]
    assert dataset.initial_concentration == "inputD1"
    assert dataset.irf == "irf1"
    assert dataset.scale == "1"

    assert "dataset2" in model.dataset
    dataset = model.dataset["dataset2"]
    assert isinstance(dataset, DatasetModel)
    assert dataset.label == "dataset2"
    assert dataset.megacomplex == ["cmplx2"]
    assert dataset.initial_concentration == "inputD2"
    assert dataset.irf == "irf2"
    assert dataset.scale == "2"
    assert dataset.spectral_axis_scale == 1e7
    assert dataset.spectral_axis_inverted


def test_constraints(model):
    print(model.clp_constraints)
    assert len(model.clp_constraints) == 2

    zero = model.clp_constraints[0]
    assert isinstance(zero, ZeroConstraint)
    assert zero.target == "s1"
    assert zero.interval == [[1, 100], [2, 200]]

    only = model.clp_constraints[1]
    assert isinstance(only, OnlyConstraint)
    assert only.target == "s1"
    assert only.interval == [[1, 100], [2, 200]]


def test_penalties(model):
    assert len(model.clp_penalties) == 1
    assert all(isinstance(c, EqualAreaPenalty) for c in model.clp_penalties)
    eac = model.clp_penalties[0]
    assert eac.type == "equal_area"
    assert eac.source == "s3"
    assert eac.source_intervals == [[670, 810]]
    assert eac.target == "s2"
    assert eac.target_intervals == [[670, 810]]
    assert eac.parameter == "55"
    assert eac.weight == 0.0016


def test_relations(model):
    print(model.clp_relations)
    assert len(model.clp_relations) == 1

    rel = model.clp_relations[0]

    assert rel.source == "s1"
    assert rel.target == "s2"
    assert rel.interval == [[1, 100], [2, 200]]


def test_initial_concentration(model):
    assert len(model.initial_concentration) == 2

    for i, _ in enumerate(model.initial_concentration, start=1):
        label = f"inputD{i}"
        assert label in model.initial_concentration
        initial_concentration = model.initial_concentration[label]
        assert initial_concentration.compartments == ["s1", "s2", "s3"]
        assert isinstance(initial_concentration, InitialConcentration)
        assert initial_concentration.label == label
        assert initial_concentration.parameters == ["1", "2", "3"]


def test_irf(model):
    assert len(model.irf) == 2

    for i, _ in enumerate(model.irf, start=1):
        label = f"irf{i}"
        assert label in model.irf
        irf = model.irf[label]
        assert isinstance(irf, IrfMultiGaussian)
        assert irf.label == label
        want = ["1"] if i == 1 else ["1", "2"]
        assert irf.center == want
        want = ["2"] if i == 1 else ["3", "4"]
        assert irf.width == want

        if i == 2:
            assert irf.center_dispersion_coefficients == ["5", "6"]
            assert irf.width_dispersion_coefficients == ["7", "8"]
            assert irf.scale == ["9"]
        assert irf.normalize == (i == 1)

        if i == 2:
            assert irf.backsweep
            assert irf.backsweep_period, 55
        else:
            assert not irf.backsweep
            assert irf.backsweep_period is None


def test_k_matrices(model):
    assert "km1" in model.k_matrix
    assert model.k_matrix["km1"].matrix == {
        ("s1", "s1"): "1",
        ("s2", "s1"): "2",
        ("s1", "s2"): "3",
        ("s3", "s1"): "4",
        ("s1", "s3"): "5",
        ("s4", "s1"): "6",
        ("s1", "s4"): "7",
    }


def test_weight(model):
    weight = model.weights[0]
    assert isinstance(weight, Weight)
    assert weight.datasets == ["d1", "d2"]
    assert weight.global_interval == [100, 102]
    assert weight.model_interval == [301, 502]
    assert weight.value == 42


def test_shapes(model):
    assert "shape1" in model.shape

    shape = model.shape["shape1"]
    assert isinstance(shape, SpectralShapeGaussian)
    assert shape.amplitude == "shape.1"
    assert shape.location == "shape.2"
    assert shape.width == "shape.3"


def test_megacomplexes(model):
    assert len(model.megacomplex) == 3

    for i in range(1, 3):
        label = f"cmplx{i}"
        assert label in model.megacomplex
        megacomplex = model.megacomplex[label]
        assert isinstance(megacomplex, DecayMegacomplex)
        assert megacomplex.label == label
        assert megacomplex.k_matrix == [f"km{i}"]

    assert "cmplx3" in model.megacomplex
    megacomplex = model.megacomplex["cmplx3"]
    assert len(megacomplex.shape) == 2
    assert megacomplex.shape["s1"] == "shape1"
    assert megacomplex.shape["s2"] == "shape2"



""" K-Matrix """
from __future__ import annotations

import itertools
from collections import OrderedDict

import numpy as np
from scipy.linalg import eig
from scipy.linalg import solve

from glotaran.model import ModelItem
from glotaran.model import ParameterType
from glotaran.model import item
from glotaran.utils.ipython import MarkdownStr


def calculate_gamma(eigenvectors: np.ndarray, initial_concentration: np.ndarray) -> np.ndarray:
    return np.diag(solve(eigenvectors, initial_concentration))


@item
class KMatrix(ModelItem):
    """A K-Matrix represents a first order differental system."""

    matrix: dict[tuple[str, str], ParameterType]

    @classmethod
    def empty(cls, label: str, compartments: list[str]) -> KMatrix:
        """Creates an empty K-Matrix. Useful for combining.

        Parameters
        ----------
        label :
            Label of the K-Matrix

        compartments :
            A list of all compartments in the model.
        """
        return cls(label, OrderedDict())

    def involved_compartments(self) -> list[str]:
        """A list of all compartments in the Matrix."""
        # TODO: find a better way that preserves ordering as defined in initial_concentrations
        compartments = []
        for index in self.matrix:
            if index[0] not in compartments:
                compartments.append(index[0])
            if index[1] not in compartments:
                compartments.append(index[1])

        # Don't use set, it randomly reorders the compartments.
        # compartments = list(set(compartments))
        return compartments

    def combine(self, k_matrix: KMatrix) -> KMatrix:
        """Creates a combined matrix.

        When combining k-matrices km1 and km2 (km1.combine(km2)),
        entries in km1 will be overwritten by corresponding entries in km2.

        Parameters
        ----------
        k_matrix :
            KMatrix to combine with.

        Returns
        -------
        combined :
            The combined KMatrix.

        """
        if not isinstance(k_matrix, KMatrix):
            raise TypeError("K-Matrices can only be combined with other K-Matrices.")
        combined_matrix = {entry: self.matrix[entry] for entry in self.matrix}
        for entry in k_matrix.matrix:
            combined_matrix[entry] = k_matrix.matrix[entry]
        return KMatrix(label=f"{self.label}+{k_matrix.label}", matrix=combined_matrix)

    def matrix_as_markdown(
        self,
        compartments: list[str] = None,
        fill_parameters: bool = False,
    ) -> MarkdownStr:
        """Returns the KMatrix as markdown formatted table.

        Parameters
        ----------
        compartments :
            (default = None)
            An optional list defining the desired order of compartments.
        fill_parameters : bool
            (default = False)
            If true, the entries will be filled with the actual parameter values
            instead of labels.
        """
        compartments = compartments or self.involved_compartments()
        size = len(compartments)
        array = np.zeros((size, size), dtype=object)
        # Matrix is a dict
        for index in self.matrix:
            i = compartments.index(index[0])
            j = compartments.index(index[1])
            array[i, j] = self.matrix[index].value if fill_parameters else self.matrix[index]

        return self._array_as_markdown(array, compartments, compartments)

    def _repr_markdown_(self) -> str:
        """Special method used by ``ipython`` to render markdown."""
        return str(self.matrix_as_markdown())

    def a_matrix_as_markdown(
        self, compartments: list[str], initial_concentration: np.ndarray
    ) -> MarkdownStr:
        """Returns the A Matrix as markdown formatted table.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        return self._array_as_markdown(
            self.a_matrix(compartments, initial_concentration).T,
            compartments,
            self.rates(compartments, initial_concentration),
        )

    @staticmethod
    def _array_as_markdown(array, row_header, column_header) -> MarkdownStr:
        markdown = "| compartment | " + " | ".join(
            e if isinstance(e, str) else f"{e:.4e}" for e in column_header
        )

        markdown += "\n|"
        markdown += "|".join("---" for _ in range(len(column_header) + 1))
        markdown += "\n"

        for i, row in enumerate(array):
            markdown += (
                f"| {row_header[i]} | "
                if isinstance(row_header[i], str)
                else f"| {row_header[i]:.4e} | "
            )
            markdown += " | ".join(e if isinstance(e, str) else f"{e:.4e}" for e in row)

            markdown += "|\n"

        return MarkdownStr(markdown)

    def reduced(self, compartments: list[str]) -> np.ndarray:
        """The reduced representation of the KMatrix as numpy array.

        Parameters
        ----------
        compartments :
            The compartment order.
        """

        size = len(compartments)
        array = np.zeros((size, size), dtype=np.float64)
        # Matrix is a dict
        for index in self.matrix:
            i = compartments.index(index[0])
            j = compartments.index(index[1])
            array[i, j] = self.matrix[index]
        return array

    def full(self, compartments: list[str]) -> np.ndarray:
        """The full representation of the KMatrix as numpy array.

        Parameters
        ----------
        compartments :
            The compartment order.
        """
        size = len(compartments)
        mat = np.zeros((size, size), np.float64)
        for (to_comp, from_comp), param in self.matrix.items():
            to_idx = compartments.index(to_comp)
            fr_idx = compartments.index(from_comp)

            if to_idx == fr_idx:
                mat[to_idx, fr_idx] -= param
            else:
                mat[to_idx, fr_idx] += param
                mat[fr_idx, fr_idx] -= param
        return mat

    def eigen(self, compartments: list[str]) -> tuple[np.ndarray, np.ndarray]:
        """Returns the eigenvalues and eigenvectors of the k matrix.

        Parameters
        ----------
        compartments :
            The compartment order.
        """
        # We take the transpose to be consistent with timp
        matrix = self.full(compartments).T
        # get the eigenvectors and values, we take the left ones to have
        # computation consistent with TIMP
        eigenvalues, eigenvectors = eig(matrix, left=True, right=False)
        return (eigenvalues.real, eigenvectors.real)

    def rates(self, compartments: list[str], initial_concentration: np.ndarray) -> np.ndarray:
        """The resulting rates of the matrix.

        By definition, the eigenvalues of the compartmental model are negative and
        the rates are the negatives of the eigenvalues, thus the eigenvalues need to be
        multiplied with ``-1`` to get rates with the correct sign.

        Parameters
        ----------
        compartments: list[str]
            Names of compartment used to order the matrix.
        initial_concentration: np.ndarray
            The initial concentration.
        """
        if self.is_sequential(compartments, initial_concentration):
            return -np.diag(self.full(compartments)).copy()
        eigenvalues, _ = self.eigen(compartments)
        return -eigenvalues

    def a_matrix(self, compartments: list[str], initial_concentration: np.ndarray) -> np.ndarray:
        """The A matrix of the KMatrix.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        return (
            self.a_matrix_sequential(compartments)
            if self.is_sequential(compartments, initial_concentration)
            else self.a_matrix_general(compartments, initial_concentration)
        )

    def a_matrix_general(
        self, compartments: list[str], initial_concentration: np.ndarray
    ) -> np.ndarray:
        """The A matrix of the KMatrix for a general model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        _, eigenvectors = self.eigen(compartments)

        gamma = calculate_gamma(eigenvectors, initial_concentration)

        a_matrix = eigenvectors @ gamma

        return a_matrix.T

    def a_matrix_sequential(self, compartments: list[str]) -> np.ndarray:
        """The A matrix of the KMatrix for a sequential model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        matrix = self.full(compartments).T
        rates = np.diag(matrix)

        a_matrix = np.zeros(matrix.shape, dtype=np.float64)
        a_matrix[0, 0] = 1.0
        for i, j in itertools.product(range(rates.size), range(1, rates.size)):
            if i > j:
                continue
            a_matrix[i, j] = np.prod([rates[m] for m in range(j)]) / np.prod(
                [rates[m] - rates[i] for m in range(j + 1) if i != m]
            )

        return a_matrix

    def is_sequential(self, compartments: list[str], initial_concentration: np.ndarray) -> bool:
        """Returns true in the KMatrix represents an unibranched model.

        Parameters
        ----------
        initial_concentration :
            The initial concentration.
        """
        if np.sum(initial_concentration) != 1:
            return False
        matrix = self.reduced(compartments)
        return not any(
            np.nonzero(matrix[:, i])[0].size != 1 or i != 0 and matrix[i, i - 1] == 0
            for i in range(matrix.shape[1])
        )
